from .repository import (
    ModelRepository,
    SchemalessModelRepository,
    DEFAULT_MODEL_SETTINGS_FILENAME,
)

from .factory import ModelRepositoryFactory

__all__ = [
    "ModelRepository",
    "SchemalessModelRepository",
    "DEFAULT_MODEL_SETTINGS_FILENAME",
    "ModelRepositoryFactory",
]



from .settings import ModelSettings
from .types import InferenceRequest, InferenceResponse


class InferenceMiddleware:
    """
    Base class to implement middlewares.
    """

    def request_middleware(
        self, request: InferenceRequest, model_settings: ModelSettings
    ) -> InferenceRequest:
        raise NotImplementedError()

    def response_middleware(
        self, response: InferenceResponse, model_settings: ModelSettings
    ) -> InferenceResponse:
        raise NotImplementedError()


class InferenceMiddlewares(InferenceMiddleware):
    """
    Meta-middleware which applies a list of middlewares.
    """

    def __init__(self, *inference_middlewares):
        self._middlewares = inference_middlewares

    def request_middleware(
        self, request: InferenceRequest, model_settings: ModelSettings
    ) -> InferenceRequest:
        processed_request = request
        for middleware in self._middlewares:
            processed_request = middleware.request_middleware(
                processed_request, model_settings
            )

        return processed_request

    def response_middleware(
        self, response: InferenceResponse, model_settings: ModelSettings
    ) -> InferenceResponse:
        processed_response = response
        for middleware in self._middlewares:
            processed_response = middleware.response_middleware(
                processed_response, model_settings
            )

        return processed_response



from collections import defaultdict, OrderedDict
from typing import Dict, List, Optional, Union, Any, DefaultDict

from ..types import (
    InferenceRequest,
    InferenceResponse,
    Parameters,
    RequestInput,
    RequestOutput,
    ResponseOutput,
)
from .shape import Shape


def _get_data(payload: Union[RequestInput, ResponseOutput]):
    return getattr(payload.data, "__root__", payload.data)


def _get_parameters(payload: ResponseOutput) -> DefaultDict[Any, Any]:
    parameters = defaultdict(list)
    if payload.parameters is not None:
        payload_parameters = payload.parameters.dict()
    for param_name, param_values in payload_parameters.items():
        if param_name in ["content_type", "headers"]:
            continue
        for param_value in param_values:
            parameters[param_name].append(param_value)
    if "content_type" in payload_parameters.keys():
        parameters["content_type"] = payload_parameters["content_type"]
    if "headers" in payload_parameters.keys():
        parameters["headers"] = payload_parameters["headers"]
    return parameters


def _merge_parameters(
    all_params: dict,
    parametrised_obj: Union[
        InferenceRequest, InferenceResponse, RequestInput, RequestOutput
    ],
) -> dict:
    if not parametrised_obj.parameters:
        return all_params

    obj_params = parametrised_obj.parameters.dict()
    return {**all_params, **obj_params}


def _merge_input_parameters(
    all_params: dict,
    parametrised_obj: Union[
        InferenceRequest, InferenceResponse, RequestInput, RequestOutput
    ],
) -> dict:
    if not parametrised_obj.parameters:
        return all_params
    obj_params = parametrised_obj.parameters.dict()
    if all_params == {}:
        return obj_params
    else:
        common_keys = set(all_params).intersection(set(obj_params)) - {
            "content_type",
            "headers",
        }
        uncommon_keys = set(all_params).union(set(obj_params)) - common_keys
        new_all_params = {}
        for key in common_keys:
            if type(all_params[key]) == list:
                new_value = all_params[key] + [obj_params[key]]
                new_all_params[key] = new_value
            else:
                new_value = [all_params[key]]
                new_value.append(obj_params[key])
                new_all_params[key] = new_value
        for key in uncommon_keys:
            if key in all_params.keys():
                new_all_params[key] = all_params[key]
            if key in obj_params.keys():
                new_all_params[key] = obj_params[key]
    return new_all_params


def _merge_data(
    all_data: Union[list, List[str], List[bytes]]
) -> Union[list, str, bytes]:
    sampled_datum = all_data[0]

    if isinstance(sampled_datum, str):
        return "".join(all_data)  # type: ignore

    if isinstance(sampled_datum, bytes):
        return b"".join(all_data)  # type: ignore

    if isinstance(sampled_datum, list):
        return sum(all_data, [])

    # TODO: Should we raise an error if we couldn't merge the data?
    return all_data


class BatchedRequests:
    def __init__(self, inference_requests: Dict[str, InferenceRequest] = {}):
        self.inference_requests = inference_requests

        # External IDs represent the incoming prediction IDs that need to match
        # 1:1 between request and response.
        # Since we can't ensure the uniqueness (or even presence) of the
        # external IDs, we'll also maintain our own list of internal IDs.
        self._ids_mapping: Dict[str, Optional[str]] = OrderedDict()

        # Minibatch here refers to the individual batch size of the input head
        # of each input request (i.e. the number of datapoints on each input
        # request)
        self._minibatch_sizes: Dict[str, int] = OrderedDict()

        self.merged_request = self._merge_requests()

    def _merge_requests(self) -> InferenceRequest:
        inputs_index: Dict[str, Dict[str, RequestInput]] = defaultdict(OrderedDict)
        outputs_index: Dict[str, Dict[str, RequestOutput]] = defaultdict(OrderedDict)
        all_params: dict = {}
        has_outputs = False  # if no outputs are defined, then outputs=None

        for internal_id, inference_request in self.inference_requests.items():
            self._ids_mapping[internal_id] = inference_request.id
            all_params = _merge_parameters(all_params, inference_request)
            for request_input in inference_request.inputs:
                inputs_index[request_input.name][internal_id] = request_input

            if inference_request.outputs is not None:
                has_outputs = True
                for request_output in inference_request.outputs:
                    outputs_index[request_output.name][internal_id] = request_output

        inputs = [
            self._merge_request_inputs(request_inputs)
            for request_inputs in inputs_index.values()
        ]

        outputs = (
            [
                self._merge_request_outputs(request_outputs)
                for request_outputs in outputs_index.values()
            ]
            if has_outputs
            else None
        )

        # TODO: Should we add a 'fake' request ID?
        params = Parameters(**all_params) if all_params else None
        return InferenceRequest(inputs=inputs, outputs=outputs, parameters=params)

    def _merge_request_inputs(
        self, request_inputs: Dict[str, RequestInput]
    ) -> RequestInput:
        # Note that minibatch sizes could be different on each input head,
        # however, to simplify the implementation, here we assume that it will
        # be the same across all of them
        batch_size = 0
        all_data = []
        all_params: dict = {}
        for internal_id, request_input in request_inputs.items():
            all_params = _merge_input_parameters(all_params, request_input)
            all_data.append(_get_data(request_input))
            minibatch_shape = Shape(request_input.shape)
            self._minibatch_sizes[internal_id] = minibatch_shape.batch_size
            batch_size += minibatch_shape.batch_size

        data = _merge_data(all_data)
        parameters = Parameters(**all_params) if all_params else None

        # TODO: What should we do if list is empty?
        sampled = next(iter(request_inputs.values()))
        shape = Shape(sampled.shape)
        shape.batch_size = batch_size

        return RequestInput(
            name=sampled.name,
            datatype=sampled.datatype,
            shape=shape.to_list(),
            data=data,
            parameters=parameters,
        )

    def _merge_request_outputs(
        self, request_outputs: Dict[str, RequestOutput]
    ) -> RequestOutput:
        all_params: dict = {}
        for internal_id, request_output in request_outputs.items():
            all_params = _merge_parameters(all_params, request_output)

        parameters = Parameters(**all_params) if all_params else None

        # TODO: What should we do if list is empty?
        sampled = next(iter(request_outputs.values()))

        return RequestOutput(name=sampled.name, parameters=parameters)

    def split_response(
        self, batched_response: InferenceResponse
    ) -> Dict[str, InferenceResponse]:
        responses: Dict[str, InferenceResponse] = {}

        for response_output in batched_response.outputs:
            response_outputs = self._split_response_output(response_output)

            for internal_id, response_output in response_outputs.items():
                if internal_id not in responses:
                    responses[internal_id] = InferenceResponse(
                        id=self._ids_mapping[internal_id],
                        model_name=batched_response.model_name,
                        model_version=batched_response.model_version,
                        outputs=[],
                        parameters=batched_response.parameters,
                    )

                responses[internal_id].outputs.append(response_output)

        return responses

    def _split_response_output(
        self, response_output: ResponseOutput
    ) -> Dict[str, ResponseOutput]:
        all_data = self._split_data(response_output)
        if response_output.parameters is not None:
            all_parameters = self._split_parameters(response_output)
        else:
            all_parameters = None
        response_outputs = {}
        for internal_id, data in all_data.items():
            shape = Shape(response_output.shape)
            shape.batch_size = self._minibatch_sizes[internal_id]
            response_outputs[internal_id] = ResponseOutput(
                name=response_output.name,
                shape=shape.to_list(),
                data=data,
                datatype=response_output.datatype,
                parameters=all_parameters
                if all_parameters is None
                else all_parameters[internal_id],
            )

        return response_outputs

    def _split_data(self, response_output: ResponseOutput) -> Dict[str, Any]:
        merged_shape = Shape(response_output.shape)
        element_size = merged_shape.elem_size
        merged_data = _get_data(response_output)
        idx = 0

        all_data = {}
        # TODO: Don't rely on array to have been flattened
        for internal_id, minibatch_size in self._minibatch_sizes.items():
            data = merged_data[idx : idx + minibatch_size * element_size]
            idx += minibatch_size * element_size
            all_data[internal_id] = data

        return all_data

    def _split_parameters(
        self, response_output: ResponseOutput
    ) -> Dict[str, Parameters]:
        merged_parameters = _get_parameters(response_output)
        idx = 0

        all_parameters = {}
        # TODO: Don't rely on array to have been flattened
        for internal_id, minibatch_size in self._minibatch_sizes.items():
            parameter_args = {}
            for parameter_name, parameter_values in merged_parameters.items():
                if parameter_name in ["content_type", "headers"]:
                    continue
                try:
                    parameter_value = parameter_values[idx]
                    if parameter_value != []:
                        parameter_args[parameter_name] = str(parameter_value)
                except IndexError:
                    pass
            if "content_type" in merged_parameters.keys():
                parameter_args["content_type"] = merged_parameters["content_type"]
            if "headers" in merged_parameters.keys():
                parameter_args["headers"] = merged_parameters["headers"]
            parameter_obj = Parameters(**parameter_args)
            all_parameters[internal_id] = parameter_obj
            idx += minibatch_size

        return all_parameters



"""
Utils to work transparently with either lists of strings, or single strings.
"""

from typing import Any, Type, Union, List, Iterator

ListElement = Union[bytes, str]
ListPayload = Union[ListElement, List[ListElement]]


def is_list_of(payload: Any, instance_type: Type):
    if not isinstance(payload, list):
        return False

    def isinstance_of_type(payload: Any) -> bool:
        return isinstance(payload, instance_type)

    return all(map(isinstance_of_type, payload))


def as_list(payload: ListPayload) -> Iterator[ListElement]:
    if isinstance(payload, list):
        # If it's a list, assume list of strings
        yield from payload
    else:
        # If not a list, assume that it's a single element
        yield payload



import asyncio

from typing import Dict, List, Tuple
from itertools import cycle
from multiprocessing import Queue
from concurrent.futures import ThreadPoolExecutor
from asyncio import Future

from ..utils import schedule_with_callback, generate_uuid

from .worker import Worker
from .logging import logger
from .utils import END_OF_QUEUE, cancel_task
from .messages import (
    Message,
    ModelUpdateMessage,
    ModelRequestMessage,
    ModelResponseMessage,
)
from prometheus_client import Histogram


class Dispatcher:
    def __init__(self, workers: Dict[int, Worker], responses: Queue):
        self._responses = responses
        self._workers = workers
        self._workers_round_robin = cycle(self._workers.keys())
        self._active = False
        self._process_responses_task = None
        self._executor = ThreadPoolExecutor()
        self._async_responses: Dict[str, Future[ModelResponseMessage]] = {}
        self.parallel_request_queue_size = Histogram(
            "parallel_request_queue",
            "counter of request queue size for workers",
            ["workerpid"],
        )

    def start(self):
        self._active = True
        self._process_responses_task = schedule_with_callback(
            self._process_responses(), self._process_responses_cb
        )

    def _process_responses_cb(self, process_responses):
        try:
            process_responses.result()
        except asyncio.CancelledError:
            # NOTE: The response loop was cancelled from the outside, so don't
            # restart
            return
        except Exception:
            logger.exception("Response processing loop crashed. Restarting the loop...")
            # If process loop crashed, restart it
            self.start()

    async def _process_responses(self):
        logger.debug("Starting response processing loop...")
        loop = asyncio.get_event_loop()
        while self._active:
            response = await loop.run_in_executor(self._executor, self._responses.get)

            # If the queue gets terminated, detect the "sentinel value" and
            # stop reading
            if response is END_OF_QUEUE:
                return

            await self._process_response(response)

    async def _process_response(self, response: ModelResponseMessage):
        internal_id = response.id

        async_response = self._async_responses[internal_id]

        # NOTE: Use call_soon_threadsafe to cover cases where `model.predict()`
        # (or other methods) get called from a separate thread (and a separate
        # AsyncIO loop)
        response_loop = async_response.get_loop()
        if response.exception:
            response_loop.call_soon_threadsafe(
                async_response.set_exception, response.exception
            )
        else:
            response_loop.call_soon_threadsafe(async_response.set_result, response)

    async def dispatch_request(
        self, request_message: ModelRequestMessage
    ) -> ModelResponseMessage:
        worker, wpid = self._get_worker()
        self._workers_queue_monitor(worker, wpid)
        worker.send_request(request_message)

        return await self._dispatch(request_message)

    def _get_worker(self) -> Tuple[Worker, int]:
        """
        Get next available worker.
        By default, this is just a round-robin through all the workers.
        """
        worker_pid = next(self._workers_round_robin)
        return self._workers[worker_pid], worker_pid

    def _workers_queue_monitor(self, worker: Worker, worker_pid: int):
        """Get metrics from every worker request queue"""
        queue_size = worker._requests.qsize()

        self.parallel_request_queue_size.labels(workerpid=str(worker_pid)).observe(
            float(queue_size)
        )

    async def dispatch_update(
        self, model_update: ModelUpdateMessage
    ) -> List[ModelResponseMessage]:
        return await asyncio.gather(
            *[
                self._dispatch_update(worker, model_update)
                for worker in self._workers.values()
            ]
        )

    async def _dispatch_update(
        self, worker: Worker, model_update: ModelUpdateMessage
    ) -> ModelResponseMessage:
        # NOTE: Need to rewrite the UUID to ensure each worker sends back a
        # unique result
        worker_update = model_update.copy()
        worker_update.id = generate_uuid()
        worker.send_update(worker_update)
        return await self._dispatch(worker_update)

    async def _dispatch(self, message: Message) -> ModelResponseMessage:
        loop = asyncio.get_running_loop()
        async_response = loop.create_future()
        internal_id = message.id
        self._async_responses[internal_id] = async_response

        return await self._wait_response(internal_id)

    async def _wait_response(self, internal_id: str) -> ModelResponseMessage:
        async_response = self._async_responses[internal_id]

        try:
            inference_response = await async_response
            return inference_response
        finally:
            del self._async_responses[internal_id]

        return await async_response

    async def stop(self):
        self._executor.shutdown()
        if self._process_responses_task is not None:
            await cancel_task(self._process_responses_task)



import pandas as pd
import numpy as np

from typing import Optional, Any, List

from .base import RequestCodec, register_request_codec
from .numpy import to_datatype, to_dtype
from .string import encode_str
from .utils import get_decoded_or_raw, InputOrOutput, inject_batch_dimension
from .lists import ListElement
from ..types import InferenceRequest, InferenceResponse, RequestInput, ResponseOutput


def _to_series(input_or_output: InputOrOutput) -> pd.Series:
    payload = get_decoded_or_raw(input_or_output)

    if input_or_output.datatype == "BYTES":
        # Don't convert the dtype of BYTES
        return pd.Series(payload)

    if isinstance(payload, np.ndarray):
        # Necessary so that it's compatible with pd.Series
        payload = list(payload)

    dtype = to_dtype(input_or_output)
    return pd.Series(payload, dtype=dtype)


def _to_response_output(series: pd.Series, use_bytes: bool = True) -> ResponseOutput:
    datatype = to_datatype(series.dtype)
    data = series.tolist()

    if datatype == "BYTES" and use_bytes:
        # To ensure that "string" columns can be encoded in gRPC, we need to
        # encode them as bytes
        data = list(map(_ensure_bytes, data))

    shape = inject_batch_dimension(list(series.shape))

    return ResponseOutput(
        name=series.name,
        shape=shape,
        # If string, it should be encoded to bytes
        data=data,
        datatype=datatype,
    )


def _ensure_bytes(elem: ListElement) -> bytes:
    if isinstance(elem, str):
        return encode_str(elem)

    return elem


@register_request_codec
class PandasCodec(RequestCodec):
    """
    Decodes a request (response) into a Pandas DataFrame, assuming each input
    (output) head corresponds to a column of the DataFrame.
    """

    ContentType = "pd"
    TypeHint = pd.DataFrame

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return isinstance(payload, pd.DataFrame)

    @classmethod
    def encode_response(
        cls,
        model_name: str,
        payload: pd.DataFrame,
        model_version: Optional[str] = None,
        use_bytes: bool = True,
        **kwargs
    ) -> InferenceResponse:
        outputs = cls.encode_outputs(payload, use_bytes=use_bytes)

        return InferenceResponse(
            model_name=model_name, model_version=model_version, outputs=outputs
        )

    @classmethod
    def decode_response(cls, response: InferenceResponse) -> pd.DataFrame:
        data = {
            response_output.name: _to_series(response_output)
            for response_output in response.outputs
        }

        return pd.DataFrame(data)

    @classmethod
    def encode_outputs(
        cls, payload: pd.DataFrame, use_bytes: bool = True
    ) -> List[ResponseOutput]:
        return [
            _to_response_output(payload[col], use_bytes=use_bytes) for col in payload
        ]

    @classmethod
    def encode_request(
        cls, payload: pd.DataFrame, use_bytes: bool = True, **kwargs
    ) -> InferenceRequest:
        outputs = cls.encode_outputs(payload, use_bytes=use_bytes)

        return InferenceRequest(
            inputs=[
                RequestInput(
                    name=output.name,
                    datatype=output.datatype,
                    shape=output.shape,
                    data=output.data,
                )
                for output in outputs
            ]
        )

    @classmethod
    def decode_request(cls, request: InferenceRequest) -> pd.DataFrame:
        data = {
            request_input.name: _to_series(request_input)
            for request_input in request.inputs
        }

        return pd.DataFrame(data)



import json

from asyncio import CancelledError
from enum import IntEnum
from pydantic import BaseModel, Field
from typing import Any, Dict, List, Optional, Union

from ..utils import get_import_path, generate_uuid
from ..settings import ModelSettings


class ModelUpdateType(IntEnum):
    Load = 1
    Unload = 2


class Message(BaseModel):
    id: str = Field(default_factory=generate_uuid)


class ModelRequestMessage(Message):
    model_name: str
    model_version: Optional[str] = None
    method_name: str
    method_args: List[Any] = []
    method_kwargs: Dict[str, Any] = {}


class ModelResponseMessage(Message):
    class Config:
        # This is to allow having an Exception field
        arbitrary_types_allowed = True

    return_value: Optional[Any]
    exception: Optional[Union[Exception, CancelledError]]


class ModelUpdateMessage(Message):
    update_type: ModelUpdateType
    serialised_model_settings: str

    def __init__(self, *args, **kwargs):
        model_settings = kwargs.pop("model_settings", None)
        if model_settings:
            as_dict = model_settings.dict()
            # Ensure the private `_source` attr also gets serialised
            if model_settings._source:
                as_dict["_source"] = model_settings._source

            import_path = get_import_path(model_settings.implementation)
            as_dict["implementation"] = import_path
            kwargs["serialised_model_settings"] = json.dumps(as_dict)
        return super().__init__(*args, **kwargs)

    @property
    def model_settings(self) -> ModelSettings:
        return ModelSettings.parse_raw(self.serialised_model_settings)



import uvicorn

from ..settings import Settings
from ..handlers import DataPlane, ModelRepositoryHandlers, get_custom_handlers
from ..model import MLModel

from .utils import matches
from .app import create_app
from .logging import logger, disable_health_access_logs
from typing import Optional


class _NoSignalServer(uvicorn.Server):
    def install_signal_handlers(self):
        pass


class RESTServer:
    def __init__(
        self,
        settings: Settings,
        data_plane: DataPlane,
        model_repository_handlers: ModelRepositoryHandlers,
    ):
        self._settings = settings
        self._data_plane = data_plane
        self._model_repository_handlers = model_repository_handlers
        self._app = create_app(
            self._settings,
            data_plane=self._data_plane,
            model_repository_handlers=self._model_repository_handlers,
        )

    async def add_custom_handlers(self, model: MLModel) -> MLModel:
        handlers = get_custom_handlers(model)
        for custom_handler, handler_method in handlers:
            self._app.add_api_route(
                custom_handler.rest_path,
                handler_method,
                methods=[custom_handler.rest_method],
            )

        return model

    async def delete_custom_handlers(self, model: MLModel) -> MLModel:
        handlers = get_custom_handlers(model)
        if len(handlers) == 0:
            return model

        # NOTE: Loop in reverse, so that it's quicker to find all the recently
        # added routes and we can remove routes on-the-fly
        for i, route in reversed(list(enumerate(self._app.routes))):
            for j, (custom_handler, handler_method) in enumerate(handlers):
                if matches(route, custom_handler, handler_method):  # type: ignore
                    self._app.routes.pop(i)
                    handlers.pop(j)

        return model

    async def start(self):
        cfg = self._get_config()
        self._server = _NoSignalServer(cfg)
        if not self._settings.debug:
            disable_health_access_logs()

        logger.info(
            "HTTP server running on "
            f"http://{self._settings.host}:{self._settings.http_port}"
        )

        await self._server.serve()

    def _get_config(self):
        kwargs = {}

        if self._settings._custom_rest_server_settings:
            logger.warning(
                "REST custom configuration is out of support. Use as your own risk"
            )
            kwargs.update(self._settings._custom_rest_server_settings)

        kwargs.update(
            {
                "host": self._settings.host,
                "port": self._settings.http_port,
                "root_path": self._settings.root_path,
                "access_log": self._settings.debug,
            }
        )

        if self._settings.logging_settings:
            # If not None, use ours. Otherwise, let Uvicorn fall back on its
            # own config.
            kwargs.update({"log_config": self._settings.logging_settings})

        return uvicorn.Config(self._app, **kwargs)

    async def stop(self, sig: Optional[int] = None):
        self._server.handle_exit(sig=sig, frame=None)



from typing import Any, Union, Mapping, Optional

from . import dataplane_pb2 as pb
from . import model_repository_pb2 as mr_pb

from .. import types
from ..raw import extract_raw, inject_raw

_FIELDS = {
    "BOOL": "bool_contents",
    "UINT8": "uint_contents",
    "UINT16": "uint_contents",
    "UINT32": "uint_contents",
    "UINT64": "uint64_contents",
    "INT8": "int_contents",
    "INT16": "int_contents",
    "INT32": "int_contents",
    "INT64": "int64_contents",
    "FP16": "bytes_contents",
    "FP32": "fp32_contents",
    "FP64": "fp64_contents",
    "BYTES": "bytes_contents",
}


def _get_value(pb_object, default: Optional[Any] = None) -> Any:
    fields = pb_object.ListFields()
    if len(fields) == 0:
        return default

    _, field_value = fields[0]
    return field_value


def _merge_map(pb_map: Mapping, value_dict: Mapping) -> Mapping:
    for key, value in value_dict.items():
        pb_map[key].MergeFrom(value)

    return pb_map


class ServerMetadataResponseConverter:
    @classmethod
    def to_types(
        cls, pb_object: pb.ServerMetadataResponse
    ) -> types.MetadataServerResponse:
        return types.MetadataServerResponse(
            name=pb_object.name,
            version=pb_object.version,
            extensions=list(pb_object.extensions),
        )

    @classmethod
    def from_types(
        cls, type_object: types.MetadataServerResponse
    ) -> pb.ServerMetadataResponse:
        return pb.ServerMetadataResponse(
            name=type_object.name,
            version=type_object.version,
            extensions=type_object.extensions,
        )


class ModelMetadataResponseConverter:
    @classmethod
    def to_types(
        cls, pb_object: pb.ModelMetadataResponse
    ) -> types.MetadataModelResponse:
        metadata = types.MetadataModelResponse(
            name=pb_object.name,
            platform=pb_object.platform,
            versions=list(pb_object.versions),
            parameters=ParametersConverter.to_types(pb_object.parameters),
        )

        if pb_object.inputs:
            metadata.inputs = [
                TensorMetadataConverter.to_types(inp) for inp in pb_object.inputs
            ]

        if pb_object.outputs:
            metadata.outputs = [
                TensorMetadataConverter.to_types(out) for out in pb_object.outputs
            ]

        return metadata

    @classmethod
    def from_types(
        cls, type_object: types.MetadataModelResponse
    ) -> pb.ModelMetadataResponse:
        metadata = pb.ModelMetadataResponse(
            name=type_object.name,
            platform=type_object.platform,
            versions=type_object.versions,
        )

        if type_object.inputs is not None:
            metadata.inputs.extend(
                [TensorMetadataConverter.from_types(inp) for inp in type_object.inputs]
            )

        if type_object.outputs is not None:
            metadata.outputs.extend(
                [TensorMetadataConverter.from_types(out) for out in type_object.outputs]
            )

        if type_object.parameters is not None:
            _merge_map(
                metadata.parameters,
                ParametersConverter.from_types(type_object.parameters),  # type: ignore
            )

        return metadata


class TensorMetadataConverter:
    @classmethod
    def to_types(
        cls, pb_object: pb.ModelMetadataResponse.TensorMetadata
    ) -> types.MetadataTensor:
        return types.MetadataTensor(
            name=pb_object.name,
            datatype=pb_object.datatype,
            shape=list(pb_object.shape),
            parameters=ParametersConverter.to_types(pb_object.parameters),
        )

    @classmethod
    def from_types(
        cls, type_object: types.MetadataTensor
    ) -> pb.ModelMetadataResponse.TensorMetadata:
        tensor_metadata = pb.ModelMetadataResponse.TensorMetadata(
            name=type_object.name,
            datatype=type_object.datatype,
            shape=type_object.shape,
        )

        if type_object.parameters is not None:
            _merge_map(
                tensor_metadata.parameters,
                ParametersConverter.from_types(type_object.parameters),  # type: ignore
            )

        return tensor_metadata


class ModelInferRequestConverter:
    @classmethod
    def to_types(cls, pb_object: pb.ModelInferRequest) -> types.InferenceRequest:
        inference_request = types.InferenceRequest.construct(
            id=pb_object.id,
            parameters=ParametersConverter.to_types(pb_object.parameters),
            inputs=[
                InferInputTensorConverter.to_types(inp) for inp in pb_object.inputs
            ],
        )

        if pb_object.outputs:
            inference_request.outputs = [
                InferRequestedOutputTensorConverter.to_types(out)
                for out in pb_object.outputs
            ]

        if pb_object.raw_input_contents:
            # Unpack and inject raw contents into `data` fields if present
            inject_raw(
                inference_request.inputs, pb_object.raw_input_contents  # type: ignore
            )

        return inference_request

    @classmethod
    def from_types(
        cls,
        type_object: types.InferenceRequest,
        model_name: str,
        model_version: str = "",
        use_raw: bool = False,
    ) -> pb.ModelInferRequest:
        if use_raw:
            # Extract the raw data in advance, to ensure the `data` field of
            # the input objects is empty
            type_object.inputs, raw = extract_raw(type_object.inputs)  # type: ignore

        model_infer_request = pb.ModelInferRequest(
            model_name=model_name,
            model_version=model_version,
            inputs=[
                InferInputTensorConverter.from_types(inp) for inp in type_object.inputs
            ],
        )

        if use_raw:
            model_infer_request.raw_input_contents.extend(raw)

        if type_object.id is not None:
            model_infer_request.id = type_object.id

        if type_object.parameters is not None:
            _merge_map(
                model_infer_request.parameters,
                ParametersConverter.from_types(type_object.parameters),
            )

        if type_object.outputs is not None:
            model_infer_request.outputs.extend(
                [
                    InferRequestedOutputTensorConverter.from_types(out)
                    for out in type_object.outputs
                ]
            )

        return model_infer_request


class InferInputTensorConverter:
    @classmethod
    def to_types(
        cls, pb_object: pb.ModelInferRequest.InferInputTensor
    ) -> types.RequestInput:
        return types.RequestInput.construct(
            name=pb_object.name,
            shape=list(pb_object.shape),
            datatype=pb_object.datatype,
            parameters=ParametersConverter.to_types(pb_object.parameters),
            data=InferTensorContentsConverter.to_types(pb_object.contents),
        )

    @classmethod
    def from_types(
        cls, type_object: types.RequestInput
    ) -> pb.ModelInferRequest.InferInputTensor:
        infer_input_tensor = pb.ModelInferRequest.InferInputTensor(
            name=type_object.name,
            shape=type_object.shape,
            datatype=type_object.datatype,
            contents=InferTensorContentsConverter.from_types(
                type_object.data, datatype=type_object.datatype
            ),
        )

        if type_object.parameters is not None:
            _merge_map(
                infer_input_tensor.parameters,
                ParametersConverter.from_types(type_object.parameters),
            )

        return infer_input_tensor


class InferRequestedOutputTensorConverter:
    @classmethod
    def to_types(
        cls, pb_object: pb.ModelInferRequest.InferRequestedOutputTensor
    ) -> types.RequestOutput:
        return types.RequestOutput.construct(
            name=pb_object.name,
            parameters=ParametersConverter.to_types(pb_object.parameters),
        )

    @classmethod
    def from_types(
        cls, type_object: types.RequestOutput
    ) -> pb.ModelInferRequest.InferRequestedOutputTensor:
        model_infer_request = pb.ModelInferRequest.InferRequestedOutputTensor(
            name=type_object.name,
        )

        if type_object.parameters is not None:
            _merge_map(
                model_infer_request.parameters,
                ParametersConverter.from_types(type_object.parameters),
            )

        return model_infer_request


class ParametersConverter:
    @classmethod
    def to_types(
        cls, pb_object: Mapping[str, pb.InferParameter]
    ) -> Optional[types.Parameters]:
        if not pb_object:
            return None

        param_dict = {
            key: _get_value(infer_parameter)
            for key, infer_parameter in pb_object.items()
        }
        return types.Parameters(**param_dict)

    @classmethod
    def from_types(
        cls, type_object: types.Parameters
    ) -> Mapping[str, pb.InferParameter]:
        pb_object = {}
        as_dict = type_object.dict()

        for key, value in as_dict.items():
            infer_parameter_key = cls._get_inferparameter_key(value)
            if infer_parameter_key is None:
                # TODO: Log warning about ignored field
                continue

            infer_parameter = pb.InferParameter(**{infer_parameter_key: value})
            pb_object[key] = infer_parameter

        return pb_object

    @classmethod
    def _get_inferparameter_key(cls, value: Union[bool, str, int]) -> Optional[str]:
        if isinstance(value, bool):
            return "bool_param"
        elif isinstance(value, str):
            return "string_param"
        elif isinstance(value, int):
            return "int64_param"

        return None


class InferTensorContentsConverter:
    @classmethod
    def to_types(cls, pb_object: pb.InferTensorContents) -> types.TensorData:
        data = _get_value(pb_object, default=[])
        as_list = list(data)
        return types.TensorData.construct(__root__=as_list)

    @classmethod
    def from_types(
        cls, type_object: types.TensorData, datatype: str
    ) -> pb.InferTensorContents:
        contents = cls._get_contents(type_object, datatype)
        return pb.InferTensorContents(**contents)

    @classmethod
    def _get_contents(cls, type_object: types.TensorData, datatype: str) -> dict:
        field = _FIELDS[datatype]
        return {field: type_object}


class ModelInferResponseConverter:
    @classmethod
    def to_types(cls, pb_object: pb.ModelInferResponse) -> types.InferenceResponse:
        inference_response = types.InferenceResponse.construct(
            id=pb_object.id,
            model_name=pb_object.model_name,
            parameters=ParametersConverter.to_types(pb_object.parameters),
            outputs=[
                InferOutputTensorConverter.to_types(output)
                for output in pb_object.outputs
            ],
        )

        if pb_object.model_version:
            inference_response.model_version = pb_object.model_version

        if pb_object.raw_output_contents:
            # Unpack and inject raw contents into `data` fields if present
            inject_raw(
                inference_response.outputs,  # type: ignore
                pb_object.raw_output_contents,  # type: ignore
            )

        return inference_response

    @classmethod
    def from_types(
        cls, type_object: types.InferenceResponse, use_raw: bool = False
    ) -> pb.ModelInferResponse:
        if use_raw:
            # Extract the raw data in advance, to ensure the `data` field of
            # the output objects is empty
            type_object.outputs, raw = extract_raw(type_object.outputs)  # type: ignore

        model_infer_response = pb.ModelInferResponse(
            model_name=type_object.model_name,
            outputs=[
                InferOutputTensorConverter.from_types(output)
                for output in type_object.outputs
            ],
        )

        if use_raw:
            # If using raw outputs, ensure it's set on the final object
            model_infer_response.raw_output_contents.extend(raw)

        if type_object.model_version is not None:
            model_infer_response.model_version = type_object.model_version

        if type_object.id is not None:
            model_infer_response.id = type_object.id

        if type_object.parameters:
            _merge_map(
                model_infer_response.parameters,
                ParametersConverter.from_types(type_object.parameters),
            )

        return model_infer_response


class InferOutputTensorConverter:
    @classmethod
    def to_types(
        cls, pb_object: pb.ModelInferResponse.InferOutputTensor
    ) -> types.ResponseOutput:
        return types.ResponseOutput(
            name=pb_object.name,
            shape=list(pb_object.shape),
            datatype=pb_object.datatype,
            parameters=ParametersConverter.to_types(pb_object.parameters),
            data=InferTensorContentsConverter.to_types(pb_object.contents),
        )

    @classmethod
    def from_types(
        cls, type_object: types.ResponseOutput
    ) -> pb.ModelInferResponse.InferOutputTensor:
        infer_output_tensor = pb.ModelInferResponse.InferOutputTensor(
            name=type_object.name,
            shape=type_object.shape,
            datatype=type_object.datatype,
            contents=InferTensorContentsConverter.from_types(
                type_object.data, datatype=type_object.datatype
            ),
        )

        if type_object.parameters:
            _merge_map(
                infer_output_tensor.parameters,
                ParametersConverter.from_types(type_object.parameters),
            )

        return infer_output_tensor


class RepositoryIndexRequestConverter:
    @classmethod
    def to_types(
        cls, pb_object: Union[pb.RepositoryIndexRequest, mr_pb.RepositoryIndexRequest]
    ) -> types.RepositoryIndexRequest:
        return types.RepositoryIndexRequest(
            ready=pb_object.ready,
        )

    @classmethod
    def from_types(
        cls, type_object: types.RepositoryIndexRequest
    ) -> Union[pb.RepositoryIndexRequest, mr_pb.RepositoryIndexRequest]:
        raise NotImplementedError("Implement me")


class RepositoryIndexResponseConverter:
    @classmethod
    def to_types(
        cls, pb_object: Union[pb.RepositoryIndexResponse, mr_pb.RepositoryIndexResponse]
    ) -> types.RepositoryIndexResponse:
        raise NotImplementedError("Implement me")

    @classmethod
    def from_types(
        cls,
        type_object: types.RepositoryIndexResponse,
        use_model_repository: bool = False,
    ) -> Union[pb.RepositoryIndexResponse, mr_pb.RepositoryIndexResponse]:
        models = [
            RepositoryIndexResponseItemConverter.from_types(
                model, use_model_repository=use_model_repository
            )
            for model in type_object
        ]
        if use_model_repository:
            return mr_pb.RepositoryIndexResponse(models=models)  # type: ignore

        return pb.RepositoryIndexResponse(models=models)  # type: ignore


class RepositoryIndexResponseItemConverter:
    @classmethod
    def to_types(
        cls,
        pb_object: Union[
            pb.RepositoryIndexResponse.ModelIndex,
            mr_pb.RepositoryIndexResponse.ModelIndex,
        ],
    ) -> types.RepositoryIndexResponseItem:
        raise NotImplementedError("Implement me")

    @classmethod
    def from_types(
        cls,
        type_object: types.RepositoryIndexResponseItem,
        use_model_repository: bool = False,
    ) -> Union[
        pb.RepositoryIndexResponse.ModelIndex, mr_pb.RepositoryIndexResponse.ModelIndex
    ]:
        model_index = pb.RepositoryIndexResponse.ModelIndex(
            name=type_object.name,
            state=type_object.state.value,
            reason=type_object.reason,
        )

        if use_model_repository:
            model_index = mr_pb.RepositoryIndexResponse.ModelIndex(  # type: ignore
                name=type_object.name,
                state=type_object.state.value,
                reason=type_object.reason,
            )

        if type_object.version is not None:
            model_index.version = type_object.version

        return model_index



import asyncio
import multiprocessing

from asyncio import Task
from multiprocessing import Queue

from ..settings import Settings

from .logging import logger

END_OF_QUEUE = None


def configure_inference_pool(settings: Settings):
    if not settings.parallel_workers:
        return

    # Set MP method
    try:
        # force set to True to override the "fork" setting used in pytest
        multiprocessing.set_start_method("spawn", force=True)
    except RuntimeError:
        # TODO: Log warning saying that mp start method couldn't be set
        method = multiprocessing.get_start_method()
        logger.exception(
            "Failed to set multiprocessing's start method. "
            f"Current method is '{method}'"
        )


async def terminate_queue(queue: Queue):
    try:
        # Send sentinel value to terminate queue
        queue.put(END_OF_QUEUE)
    except (ValueError, AssertionError):
        # Most likely, queue is already closed
        return


async def cancel_task(task: Task):
    try:
        task.cancel()
        await task
    except asyncio.CancelledError:
        pass



from typing import Any, Union, Dict, Optional, List

from ..types import (
    InferenceRequest,
    InferenceResponse,
    RequestInput,
    RequestOutput,
    ResponseOutput,
    MetadataTensor,
    Parameters,
)
from ..settings import ModelSettings
from .base import (
    find_input_codec,
    find_input_codec_by_payload,
    find_request_codec,
    find_request_codec_by_payload,
    RequestCodec,
    InputCodecLike,
    RequestCodecLike,
)
from .errors import CodecError

DefaultOutputPrefix = "output-"
DefaultInputPrefix = "input-"

InputOrOutput = Union[RequestInput, ResponseOutput]
Codec = Union[InputCodecLike, RequestCodecLike]

Parametrised = Union[
    InferenceRequest, RequestInput, RequestOutput, ResponseOutput, InferenceResponse
]
Tagged = Union[MetadataTensor, ModelSettings]
DecodedParameterName = "_decoded_payload"


def inject_batch_dimension(shape: List[int]) -> List[int]:
    """
    Utility method to ensure that 1-dimensional shapes
    assume that `[N] == [N, D]`.
    """
    if len(shape) > 1:
        return shape

    return shape + [1]


def _get_content_type(
    request: Parametrised, metadata: Optional[Tagged] = None
) -> Optional[str]:
    if request.parameters and request.parameters.content_type:
        return request.parameters.content_type

    if metadata is not None:
        if metadata.parameters and metadata.parameters.content_type:
            return metadata.parameters.content_type

    return None


def _save_decoded(parametrised_obj: Parametrised, decoded_payload: Any):
    if not parametrised_obj.parameters:
        parametrised_obj.parameters = Parameters()

    setattr(parametrised_obj.parameters, DecodedParameterName, decoded_payload)


def encode_response_output(
    payload: Any,
    request_output: RequestOutput,
    metadata_outputs: Dict[str, MetadataTensor] = {},
) -> Optional[ResponseOutput]:
    output_metadata = metadata_outputs.get(request_output.name)
    content_type = _get_content_type(request_output, output_metadata)
    codec = (
        find_input_codec(content_type)
        if content_type
        else find_input_codec_by_payload(payload)
    )

    if not codec:
        return None

    return codec.encode_output(
        name=request_output.name,
        payload=payload,
    )


def encode_inference_response(
    payload: Any,
    model_settings: ModelSettings,
) -> Optional[InferenceResponse]:
    # TODO: Allow users to override codec through model's metadata
    codec = find_request_codec_by_payload(payload)

    if not codec:
        return None

    model_version = None
    if model_settings.parameters:
        model_version = model_settings.parameters.version

    return codec.encode_response(model_settings.name, payload, model_version)


def decode_request_input(
    request_input: RequestInput,
    metadata_inputs: Dict[str, MetadataTensor] = {},
) -> Optional[Any]:
    input_metadata = metadata_inputs.get(request_input.name)
    content_type = _get_content_type(request_input, input_metadata)
    if content_type is None:
        return None

    codec = find_input_codec(content_type)
    if codec is None:
        return None

    decoded_payload = codec.decode_input(request_input)
    _save_decoded(request_input, decoded_payload)
    return decoded_payload


def decode_inference_request(
    inference_request: InferenceRequest,
    model_settings: Optional[ModelSettings] = None,
    metadata_inputs: Dict[str, MetadataTensor] = {},
) -> Optional[Any]:
    for request_input in inference_request.inputs:
        decode_request_input(request_input, metadata_inputs)

    request_content_type = _get_content_type(inference_request, model_settings)
    if request_content_type is not None:
        codec = find_request_codec(request_content_type)
        if codec is not None:
            decoded_payload = codec.decode_request(inference_request)
            _save_decoded(inference_request, decoded_payload)
            return decoded_payload

    return inference_request


def has_decoded(parametrised_obj: Parametrised) -> bool:
    if parametrised_obj.parameters:
        return hasattr(parametrised_obj.parameters, DecodedParameterName)

    return False


def get_decoded(parametrised_obj: Parametrised) -> Any:
    if has_decoded(parametrised_obj):
        return getattr(parametrised_obj.parameters, DecodedParameterName)


def get_decoded_or_raw(parametrised_obj: Parametrised) -> Any:
    if not has_decoded(parametrised_obj):
        if isinstance(parametrised_obj, RequestInput):
            # If this is a RequestInput, return its data
            return parametrised_obj.data

        if isinstance(parametrised_obj, ResponseOutput):
            # If this is a ResponseOutput, return its data
            return parametrised_obj.data

        # Otherwise, return full object
        return parametrised_obj

    return get_decoded(parametrised_obj)


class SingleInputRequestCodec(RequestCodec):
    """
    The SingleInputRequestCodec can be used as a "meta-implementation" for other
    codecs. Its goal to decode the whole request simply as the first decoded
    element.
    """

    InputCodec: Optional[InputCodecLike] = None

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        if cls.InputCodec is None:
            return False

        return cls.InputCodec.can_encode(payload)

    @classmethod
    def encode_response(
        cls,
        model_name: str,
        payload: Any,
        model_version: Optional[str] = None,
        **kwargs,
    ) -> InferenceResponse:
        if cls.InputCodec is None:
            raise NotImplementedError(
                f"No input codec found for {type(cls)} request codec"
            )

        output = cls.InputCodec.encode_output(
            f"{DefaultOutputPrefix}1", payload, **kwargs
        )
        return InferenceResponse(
            model_name=model_name, model_version=model_version, outputs=[output]
        )

    @classmethod
    def decode_response(cls, response: InferenceResponse) -> Any:
        if len(response.outputs) != 1:
            raise CodecError(
                f"The '{cls.ContentType}' codec only supports a single output tensor "
                f"({len(response.outputs)} were received)"
            )

        first_output = response.outputs[0]
        if not has_decoded(first_output) and cls.InputCodec is not None:
            decoded_payload = cls.InputCodec.decode_output(first_output)  # type: ignore
            _save_decoded(first_output, decoded_payload)

        return get_decoded_or_raw(first_output)

    @classmethod
    def encode_request(cls, payload: Any, **kwargs) -> InferenceRequest:
        if cls.InputCodec is None:
            raise NotImplementedError(
                f"No input codec found for {type(cls)} request codec"
            )

        inp = cls.InputCodec.encode_input(f"{DefaultInputPrefix}1", payload, **kwargs)
        return InferenceRequest(inputs=[inp])

    @classmethod
    def decode_request(cls, request: InferenceRequest) -> Any:
        if len(request.inputs) != 1:
            raise CodecError(
                f"The '{cls.ContentType}' codec only supports a single input tensor "
                f"({len(request.inputs)} were received)"
            )

        first_input = request.inputs[0]
        if not has_decoded(first_input) and cls.InputCodec is not None:
            decoded_payload = cls.InputCodec.decode_input(first_input)  # type: ignore
            _save_decoded(first_input, decoded_payload)

        return get_decoded_or_raw(first_input)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dataplane.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0f\x64\x61taplane.proto\x12\tinference"\x13\n\x11ServerLiveRequest""\n\x12ServerLiveResponse\x12\x0c\n\x04live\x18\x01 \x01(\x08"\x14\n\x12ServerReadyRequest"$\n\x13ServerReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"2\n\x11ModelReadyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t"#\n\x12ModelReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"\x17\n\x15ServerMetadataRequest"K\n\x16ServerMetadataResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x12\n\nextensions\x18\x03 \x03(\t"5\n\x14ModelMetadataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t"\xc5\x04\n\x15ModelMetadataResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08versions\x18\x02 \x03(\t\x12\x10\n\x08platform\x18\x03 \x01(\t\x12?\n\x06inputs\x18\x04 \x03(\x0b\x32/.inference.ModelMetadataResponse.TensorMetadata\x12@\n\x07outputs\x18\x05 \x03(\x0b\x32/.inference.ModelMetadataResponse.TensorMetadata\x12\x44\n\nparameters\x18\x06 \x03(\x0b\x32\x30.inference.ModelMetadataResponse.ParametersEntry\x1a\xe2\x01\n\x0eTensorMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\x12S\n\nparameters\x18\x04 \x03(\x0b\x32?.inference.ModelMetadataResponse.TensorMetadata.ParametersEntry\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01"\xee\x06\n\x11ModelInferRequest\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12@\n\nparameters\x18\x04 \x03(\x0b\x32,.inference.ModelInferRequest.ParametersEntry\x12=\n\x06inputs\x18\x05 \x03(\x0b\x32-.inference.ModelInferRequest.InferInputTensor\x12H\n\x07outputs\x18\x06 \x03(\x0b\x32\x37.inference.ModelInferRequest.InferRequestedOutputTensor\x12\x1a\n\x12raw_input_contents\x18\x07 \x03(\x0c\x1a\x94\x02\n\x10InferInputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\x12Q\n\nparameters\x18\x04 \x03(\x0b\x32=.inference.ModelInferRequest.InferInputTensor.ParametersEntry\x12\x30\n\x08\x63ontents\x18\x05 \x01(\x0b\x32\x1e.inference.InferTensorContents\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1a\xd5\x01\n\x1aInferRequestedOutputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12[\n\nparameters\x18\x02 \x03(\x0b\x32G.inference.ModelInferRequest.InferRequestedOutputTensor.ParametersEntry\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01"\xd5\x04\n\x12ModelInferResponse\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12\x41\n\nparameters\x18\x04 \x03(\x0b\x32-.inference.ModelInferResponse.ParametersEntry\x12@\n\x07outputs\x18\x05 \x03(\x0b\x32/.inference.ModelInferResponse.InferOutputTensor\x12\x1b\n\x13raw_output_contents\x18\x06 \x03(\x0c\x1a\x97\x02\n\x11InferOutputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\x12S\n\nparameters\x18\x04 \x03(\x0b\x32?.inference.ModelInferResponse.InferOutputTensor.ParametersEntry\x12\x30\n\x08\x63ontents\x18\x05 \x01(\x0b\x32\x1e.inference.InferTensorContents\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01"i\n\x0eInferParameter\x12\x14\n\nbool_param\x18\x01 \x01(\x08H\x00\x12\x15\n\x0bint64_param\x18\x02 \x01(\x03H\x00\x12\x16\n\x0cstring_param\x18\x03 \x01(\tH\x00\x42\x12\n\x10parameter_choice"\xd0\x01\n\x13InferTensorContents\x12\x15\n\rbool_contents\x18\x01 \x03(\x08\x12\x14\n\x0cint_contents\x18\x02 \x03(\x05\x12\x16\n\x0eint64_contents\x18\x03 \x03(\x03\x12\x15\n\ruint_contents\x18\x04 \x03(\r\x12\x17\n\x0fuint64_contents\x18\x05 \x03(\x04\x12\x15\n\rfp32_contents\x18\x06 \x03(\x02\x12\x15\n\rfp64_contents\x18\x07 \x03(\x01\x12\x16\n\x0e\x62ytes_contents\x18\x08 \x03(\x0c"\x8a\x01\n\x18ModelRepositoryParameter\x12\x14\n\nbool_param\x18\x01 \x01(\x08H\x00\x12\x15\n\x0bint64_param\x18\x02 \x01(\x03H\x00\x12\x16\n\x0cstring_param\x18\x03 \x01(\tH\x00\x12\x15\n\x0b\x62ytes_param\x18\x04 \x01(\x0cH\x00\x42\x12\n\x10parameter_choice"@\n\x16RepositoryIndexRequest\x12\x17\n\x0frepository_name\x18\x01 \x01(\t\x12\r\n\x05ready\x18\x02 \x01(\x08"\xa4\x01\n\x17RepositoryIndexResponse\x12=\n\x06models\x18\x01 \x03(\x0b\x32-.inference.RepositoryIndexResponse.ModelIndex\x1aJ\n\nModelIndex\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\r\n\x05state\x18\x03 \x01(\t\x12\x0e\n\x06reason\x18\x04 \x01(\t"\xec\x01\n\x1aRepositoryModelLoadRequest\x12\x17\n\x0frepository_name\x18\x01 \x01(\t\x12\x12\n\nmodel_name\x18\x02 \x01(\t\x12I\n\nparameters\x18\x03 \x03(\x0b\x32\x35.inference.RepositoryModelLoadRequest.ParametersEntry\x1aV\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.inference.ModelRepositoryParameter:\x02\x38\x01"\x1d\n\x1bRepositoryModelLoadResponse"\xf0\x01\n\x1cRepositoryModelUnloadRequest\x12\x17\n\x0frepository_name\x18\x01 \x01(\t\x12\x12\n\nmodel_name\x18\x02 \x01(\t\x12K\n\nparameters\x18\x03 \x03(\x0b\x32\x37.inference.RepositoryModelUnloadRequest.ParametersEntry\x1aV\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.inference.ModelRepositoryParameter:\x02\x38\x01"\x1f\n\x1dRepositoryModelUnloadResponse2\xae\x06\n\x14GRPCInferenceService\x12K\n\nServerLive\x12\x1c.inference.ServerLiveRequest\x1a\x1d.inference.ServerLiveResponse"\x00\x12N\n\x0bServerReady\x12\x1d.inference.ServerReadyRequest\x1a\x1e.inference.ServerReadyResponse"\x00\x12K\n\nModelReady\x12\x1c.inference.ModelReadyRequest\x1a\x1d.inference.ModelReadyResponse"\x00\x12W\n\x0eServerMetadata\x12 .inference.ServerMetadataRequest\x1a!.inference.ServerMetadataResponse"\x00\x12T\n\rModelMetadata\x12\x1f.inference.ModelMetadataRequest\x1a .inference.ModelMetadataResponse"\x00\x12K\n\nModelInfer\x12\x1c.inference.ModelInferRequest\x1a\x1d.inference.ModelInferResponse"\x00\x12Z\n\x0fRepositoryIndex\x12!.inference.RepositoryIndexRequest\x1a".inference.RepositoryIndexResponse"\x00\x12\x66\n\x13RepositoryModelLoad\x12%.inference.RepositoryModelLoadRequest\x1a&.inference.RepositoryModelLoadResponse"\x00\x12l\n\x15RepositoryModelUnload\x12\'.inference.RepositoryModelUnloadRequest\x1a(.inference.RepositoryModelUnloadResponse"\x00\x62\x06proto3'
)


_SERVERLIVEREQUEST = DESCRIPTOR.message_types_by_name["ServerLiveRequest"]
_SERVERLIVERESPONSE = DESCRIPTOR.message_types_by_name["ServerLiveResponse"]
_SERVERREADYREQUEST = DESCRIPTOR.message_types_by_name["ServerReadyRequest"]
_SERVERREADYRESPONSE = DESCRIPTOR.message_types_by_name["ServerReadyResponse"]
_MODELREADYREQUEST = DESCRIPTOR.message_types_by_name["ModelReadyRequest"]
_MODELREADYRESPONSE = DESCRIPTOR.message_types_by_name["ModelReadyResponse"]
_SERVERMETADATAREQUEST = DESCRIPTOR.message_types_by_name["ServerMetadataRequest"]
_SERVERMETADATARESPONSE = DESCRIPTOR.message_types_by_name["ServerMetadataResponse"]
_MODELMETADATAREQUEST = DESCRIPTOR.message_types_by_name["ModelMetadataRequest"]
_MODELMETADATARESPONSE = DESCRIPTOR.message_types_by_name["ModelMetadataResponse"]
_MODELMETADATARESPONSE_TENSORMETADATA = _MODELMETADATARESPONSE.nested_types_by_name[
    "TensorMetadata"
]
_MODELMETADATARESPONSE_TENSORMETADATA_PARAMETERSENTRY = (
    _MODELMETADATARESPONSE_TENSORMETADATA.nested_types_by_name["ParametersEntry"]
)
_MODELMETADATARESPONSE_PARAMETERSENTRY = _MODELMETADATARESPONSE.nested_types_by_name[
    "ParametersEntry"
]
_MODELINFERREQUEST = DESCRIPTOR.message_types_by_name["ModelInferRequest"]
_MODELINFERREQUEST_INFERINPUTTENSOR = _MODELINFERREQUEST.nested_types_by_name[
    "InferInputTensor"
]
_MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY = (
    _MODELINFERREQUEST_INFERINPUTTENSOR.nested_types_by_name["ParametersEntry"]
)
_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR = _MODELINFERREQUEST.nested_types_by_name[
    "InferRequestedOutputTensor"
]
_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY = (
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR.nested_types_by_name[
        "ParametersEntry"
    ]
)
_MODELINFERREQUEST_PARAMETERSENTRY = _MODELINFERREQUEST.nested_types_by_name[
    "ParametersEntry"
]
_MODELINFERRESPONSE = DESCRIPTOR.message_types_by_name["ModelInferResponse"]
_MODELINFERRESPONSE_INFEROUTPUTTENSOR = _MODELINFERRESPONSE.nested_types_by_name[
    "InferOutputTensor"
]
_MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY = (
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR.nested_types_by_name["ParametersEntry"]
)
_MODELINFERRESPONSE_PARAMETERSENTRY = _MODELINFERRESPONSE.nested_types_by_name[
    "ParametersEntry"
]
_INFERPARAMETER = DESCRIPTOR.message_types_by_name["InferParameter"]
_INFERTENSORCONTENTS = DESCRIPTOR.message_types_by_name["InferTensorContents"]
_MODELREPOSITORYPARAMETER = DESCRIPTOR.message_types_by_name["ModelRepositoryParameter"]
_REPOSITORYINDEXREQUEST = DESCRIPTOR.message_types_by_name["RepositoryIndexRequest"]
_REPOSITORYINDEXRESPONSE = DESCRIPTOR.message_types_by_name["RepositoryIndexResponse"]
_REPOSITORYINDEXRESPONSE_MODELINDEX = _REPOSITORYINDEXRESPONSE.nested_types_by_name[
    "ModelIndex"
]
_REPOSITORYMODELLOADREQUEST = DESCRIPTOR.message_types_by_name[
    "RepositoryModelLoadRequest"
]
_REPOSITORYMODELLOADREQUEST_PARAMETERSENTRY = (
    _REPOSITORYMODELLOADREQUEST.nested_types_by_name["ParametersEntry"]
)
_REPOSITORYMODELLOADRESPONSE = DESCRIPTOR.message_types_by_name[
    "RepositoryModelLoadResponse"
]
_REPOSITORYMODELUNLOADREQUEST = DESCRIPTOR.message_types_by_name[
    "RepositoryModelUnloadRequest"
]
_REPOSITORYMODELUNLOADREQUEST_PARAMETERSENTRY = (
    _REPOSITORYMODELUNLOADREQUEST.nested_types_by_name["ParametersEntry"]
)
_REPOSITORYMODELUNLOADRESPONSE = DESCRIPTOR.message_types_by_name[
    "RepositoryModelUnloadResponse"
]
ServerLiveRequest = _reflection.GeneratedProtocolMessageType(
    "ServerLiveRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERLIVEREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ServerLiveRequest)
    },
)
_sym_db.RegisterMessage(ServerLiveRequest)

ServerLiveResponse = _reflection.GeneratedProtocolMessageType(
    "ServerLiveResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERLIVERESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ServerLiveResponse)
    },
)
_sym_db.RegisterMessage(ServerLiveResponse)

ServerReadyRequest = _reflection.GeneratedProtocolMessageType(
    "ServerReadyRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERREADYREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ServerReadyRequest)
    },
)
_sym_db.RegisterMessage(ServerReadyRequest)

ServerReadyResponse = _reflection.GeneratedProtocolMessageType(
    "ServerReadyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERREADYRESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ServerReadyResponse)
    },
)
_sym_db.RegisterMessage(ServerReadyResponse)

ModelReadyRequest = _reflection.GeneratedProtocolMessageType(
    "ModelReadyRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELREADYREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ModelReadyRequest)
    },
)
_sym_db.RegisterMessage(ModelReadyRequest)

ModelReadyResponse = _reflection.GeneratedProtocolMessageType(
    "ModelReadyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELREADYRESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ModelReadyResponse)
    },
)
_sym_db.RegisterMessage(ModelReadyResponse)

ServerMetadataRequest = _reflection.GeneratedProtocolMessageType(
    "ServerMetadataRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERMETADATAREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ServerMetadataRequest)
    },
)
_sym_db.RegisterMessage(ServerMetadataRequest)

ServerMetadataResponse = _reflection.GeneratedProtocolMessageType(
    "ServerMetadataResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERMETADATARESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ServerMetadataResponse)
    },
)
_sym_db.RegisterMessage(ServerMetadataResponse)

ModelMetadataRequest = _reflection.GeneratedProtocolMessageType(
    "ModelMetadataRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELMETADATAREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ModelMetadataRequest)
    },
)
_sym_db.RegisterMessage(ModelMetadataRequest)

ModelMetadataResponse = _reflection.GeneratedProtocolMessageType(
    "ModelMetadataResponse",
    (_message.Message,),
    {
        "TensorMetadata": _reflection.GeneratedProtocolMessageType(
            "TensorMetadata",
            (_message.Message,),
            {
                "ParametersEntry": _reflection.GeneratedProtocolMessageType(
                    "ParametersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _MODELMETADATARESPONSE_TENSORMETADATA_PARAMETERSENTRY,
                        "__module__": "dataplane_pb2"
                        # @@protoc_insertion_point(class_scope:inference.ModelMetadataResponse.TensorMetadata.ParametersEntry)
                    },
                ),
                "DESCRIPTOR": _MODELMETADATARESPONSE_TENSORMETADATA,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.ModelMetadataResponse.TensorMetadata)
            },
        ),
        "ParametersEntry": _reflection.GeneratedProtocolMessageType(
            "ParametersEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _MODELMETADATARESPONSE_PARAMETERSENTRY,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.ModelMetadataResponse.ParametersEntry)
            },
        ),
        "DESCRIPTOR": _MODELMETADATARESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ModelMetadataResponse)
    },
)
_sym_db.RegisterMessage(ModelMetadataResponse)
_sym_db.RegisterMessage(ModelMetadataResponse.TensorMetadata)
_sym_db.RegisterMessage(ModelMetadataResponse.TensorMetadata.ParametersEntry)
_sym_db.RegisterMessage(ModelMetadataResponse.ParametersEntry)

ModelInferRequest = _reflection.GeneratedProtocolMessageType(
    "ModelInferRequest",
    (_message.Message,),
    {
        "InferInputTensor": _reflection.GeneratedProtocolMessageType(
            "InferInputTensor",
            (_message.Message,),
            {
                "ParametersEntry": _reflection.GeneratedProtocolMessageType(
                    "ParametersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY,
                        "__module__": "dataplane_pb2"
                        # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.InferInputTensor.ParametersEntry)
                    },
                ),
                "DESCRIPTOR": _MODELINFERREQUEST_INFERINPUTTENSOR,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.InferInputTensor)
            },
        ),
        "InferRequestedOutputTensor": _reflection.GeneratedProtocolMessageType(
            "InferRequestedOutputTensor",
            (_message.Message,),
            {
                "ParametersEntry": _reflection.GeneratedProtocolMessageType(
                    "ParametersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY,
                        "__module__": "dataplane_pb2"
                        # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.InferRequestedOutputTensor.ParametersEntry)
                    },
                ),
                "DESCRIPTOR": _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.InferRequestedOutputTensor)
            },
        ),
        "ParametersEntry": _reflection.GeneratedProtocolMessageType(
            "ParametersEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _MODELINFERREQUEST_PARAMETERSENTRY,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.ParametersEntry)
            },
        ),
        "DESCRIPTOR": _MODELINFERREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ModelInferRequest)
    },
)
_sym_db.RegisterMessage(ModelInferRequest)
_sym_db.RegisterMessage(ModelInferRequest.InferInputTensor)
_sym_db.RegisterMessage(ModelInferRequest.InferInputTensor.ParametersEntry)
_sym_db.RegisterMessage(ModelInferRequest.InferRequestedOutputTensor)
_sym_db.RegisterMessage(ModelInferRequest.InferRequestedOutputTensor.ParametersEntry)
_sym_db.RegisterMessage(ModelInferRequest.ParametersEntry)

ModelInferResponse = _reflection.GeneratedProtocolMessageType(
    "ModelInferResponse",
    (_message.Message,),
    {
        "InferOutputTensor": _reflection.GeneratedProtocolMessageType(
            "InferOutputTensor",
            (_message.Message,),
            {
                "ParametersEntry": _reflection.GeneratedProtocolMessageType(
                    "ParametersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY,
                        "__module__": "dataplane_pb2"
                        # @@protoc_insertion_point(class_scope:inference.ModelInferResponse.InferOutputTensor.ParametersEntry)
                    },
                ),
                "DESCRIPTOR": _MODELINFERRESPONSE_INFEROUTPUTTENSOR,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.ModelInferResponse.InferOutputTensor)
            },
        ),
        "ParametersEntry": _reflection.GeneratedProtocolMessageType(
            "ParametersEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _MODELINFERRESPONSE_PARAMETERSENTRY,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.ModelInferResponse.ParametersEntry)
            },
        ),
        "DESCRIPTOR": _MODELINFERRESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ModelInferResponse)
    },
)
_sym_db.RegisterMessage(ModelInferResponse)
_sym_db.RegisterMessage(ModelInferResponse.InferOutputTensor)
_sym_db.RegisterMessage(ModelInferResponse.InferOutputTensor.ParametersEntry)
_sym_db.RegisterMessage(ModelInferResponse.ParametersEntry)

InferParameter = _reflection.GeneratedProtocolMessageType(
    "InferParameter",
    (_message.Message,),
    {
        "DESCRIPTOR": _INFERPARAMETER,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.InferParameter)
    },
)
_sym_db.RegisterMessage(InferParameter)

InferTensorContents = _reflection.GeneratedProtocolMessageType(
    "InferTensorContents",
    (_message.Message,),
    {
        "DESCRIPTOR": _INFERTENSORCONTENTS,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.InferTensorContents)
    },
)
_sym_db.RegisterMessage(InferTensorContents)

ModelRepositoryParameter = _reflection.GeneratedProtocolMessageType(
    "ModelRepositoryParameter",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELREPOSITORYPARAMETER,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.ModelRepositoryParameter)
    },
)
_sym_db.RegisterMessage(ModelRepositoryParameter)

RepositoryIndexRequest = _reflection.GeneratedProtocolMessageType(
    "RepositoryIndexRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYINDEXREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.RepositoryIndexRequest)
    },
)
_sym_db.RegisterMessage(RepositoryIndexRequest)

RepositoryIndexResponse = _reflection.GeneratedProtocolMessageType(
    "RepositoryIndexResponse",
    (_message.Message,),
    {
        "ModelIndex": _reflection.GeneratedProtocolMessageType(
            "ModelIndex",
            (_message.Message,),
            {
                "DESCRIPTOR": _REPOSITORYINDEXRESPONSE_MODELINDEX,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.RepositoryIndexResponse.ModelIndex)
            },
        ),
        "DESCRIPTOR": _REPOSITORYINDEXRESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.RepositoryIndexResponse)
    },
)
_sym_db.RegisterMessage(RepositoryIndexResponse)
_sym_db.RegisterMessage(RepositoryIndexResponse.ModelIndex)

RepositoryModelLoadRequest = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelLoadRequest",
    (_message.Message,),
    {
        "ParametersEntry": _reflection.GeneratedProtocolMessageType(
            "ParametersEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _REPOSITORYMODELLOADREQUEST_PARAMETERSENTRY,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.RepositoryModelLoadRequest.ParametersEntry)
            },
        ),
        "DESCRIPTOR": _REPOSITORYMODELLOADREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.RepositoryModelLoadRequest)
    },
)
_sym_db.RegisterMessage(RepositoryModelLoadRequest)
_sym_db.RegisterMessage(RepositoryModelLoadRequest.ParametersEntry)

RepositoryModelLoadResponse = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelLoadResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELLOADRESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.RepositoryModelLoadResponse)
    },
)
_sym_db.RegisterMessage(RepositoryModelLoadResponse)

RepositoryModelUnloadRequest = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelUnloadRequest",
    (_message.Message,),
    {
        "ParametersEntry": _reflection.GeneratedProtocolMessageType(
            "ParametersEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _REPOSITORYMODELUNLOADREQUEST_PARAMETERSENTRY,
                "__module__": "dataplane_pb2"
                # @@protoc_insertion_point(class_scope:inference.RepositoryModelUnloadRequest.ParametersEntry)
            },
        ),
        "DESCRIPTOR": _REPOSITORYMODELUNLOADREQUEST,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.RepositoryModelUnloadRequest)
    },
)
_sym_db.RegisterMessage(RepositoryModelUnloadRequest)
_sym_db.RegisterMessage(RepositoryModelUnloadRequest.ParametersEntry)

RepositoryModelUnloadResponse = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelUnloadResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELUNLOADRESPONSE,
        "__module__": "dataplane_pb2"
        # @@protoc_insertion_point(class_scope:inference.RepositoryModelUnloadResponse)
    },
)
_sym_db.RegisterMessage(RepositoryModelUnloadResponse)

_GRPCINFERENCESERVICE = DESCRIPTOR.services_by_name["GRPCInferenceService"]
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _MODELMETADATARESPONSE_TENSORMETADATA_PARAMETERSENTRY._options = None
    _MODELMETADATARESPONSE_TENSORMETADATA_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELMETADATARESPONSE_PARAMETERSENTRY._options = None
    _MODELMETADATARESPONSE_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY._options = None
    _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY._options = None
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY._serialized_options = (
        b"8\001"
    )
    _MODELINFERREQUEST_PARAMETERSENTRY._options = None
    _MODELINFERREQUEST_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY._options = None
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELINFERRESPONSE_PARAMETERSENTRY._options = None
    _MODELINFERRESPONSE_PARAMETERSENTRY._serialized_options = b"8\001"
    _REPOSITORYMODELLOADREQUEST_PARAMETERSENTRY._options = None
    _REPOSITORYMODELLOADREQUEST_PARAMETERSENTRY._serialized_options = b"8\001"
    _REPOSITORYMODELUNLOADREQUEST_PARAMETERSENTRY._options = None
    _REPOSITORYMODELUNLOADREQUEST_PARAMETERSENTRY._serialized_options = b"8\001"
    _SERVERLIVEREQUEST._serialized_start = 30
    _SERVERLIVEREQUEST._serialized_end = 49
    _SERVERLIVERESPONSE._serialized_start = 51
    _SERVERLIVERESPONSE._serialized_end = 85
    _SERVERREADYREQUEST._serialized_start = 87
    _SERVERREADYREQUEST._serialized_end = 107
    _SERVERREADYRESPONSE._serialized_start = 109
    _SERVERREADYRESPONSE._serialized_end = 145
    _MODELREADYREQUEST._serialized_start = 147
    _MODELREADYREQUEST._serialized_end = 197
    _MODELREADYRESPONSE._serialized_start = 199
    _MODELREADYRESPONSE._serialized_end = 234
    _SERVERMETADATAREQUEST._serialized_start = 236
    _SERVERMETADATAREQUEST._serialized_end = 259
    _SERVERMETADATARESPONSE._serialized_start = 261
    _SERVERMETADATARESPONSE._serialized_end = 336
    _MODELMETADATAREQUEST._serialized_start = 338
    _MODELMETADATAREQUEST._serialized_end = 391
    _MODELMETADATARESPONSE._serialized_start = 394
    _MODELMETADATARESPONSE._serialized_end = 975
    _MODELMETADATARESPONSE_TENSORMETADATA._serialized_start = 671
    _MODELMETADATARESPONSE_TENSORMETADATA._serialized_end = 897
    _MODELMETADATARESPONSE_TENSORMETADATA_PARAMETERSENTRY._serialized_start = 821
    _MODELMETADATARESPONSE_TENSORMETADATA_PARAMETERSENTRY._serialized_end = 897
    _MODELMETADATARESPONSE_PARAMETERSENTRY._serialized_start = 821
    _MODELMETADATARESPONSE_PARAMETERSENTRY._serialized_end = 897
    _MODELINFERREQUEST._serialized_start = 978
    _MODELINFERREQUEST._serialized_end = 1856
    _MODELINFERREQUEST_INFERINPUTTENSOR._serialized_start = 1286
    _MODELINFERREQUEST_INFERINPUTTENSOR._serialized_end = 1562
    _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY._serialized_start = 821
    _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY._serialized_end = 897
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR._serialized_start = 1565
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR._serialized_end = 1778
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY._serialized_start = (
        821
    )
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY._serialized_end = 897
    _MODELINFERREQUEST_PARAMETERSENTRY._serialized_start = 821
    _MODELINFERREQUEST_PARAMETERSENTRY._serialized_end = 897
    _MODELINFERRESPONSE._serialized_start = 1859
    _MODELINFERRESPONSE._serialized_end = 2456
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR._serialized_start = 2099
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR._serialized_end = 2378
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY._serialized_start = 821
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY._serialized_end = 897
    _MODELINFERRESPONSE_PARAMETERSENTRY._serialized_start = 821
    _MODELINFERRESPONSE_PARAMETERSENTRY._serialized_end = 897
    _INFERPARAMETER._serialized_start = 2458
    _INFERPARAMETER._serialized_end = 2563
    _INFERTENSORCONTENTS._serialized_start = 2566
    _INFERTENSORCONTENTS._serialized_end = 2774
    _MODELREPOSITORYPARAMETER._serialized_start = 2777
    _MODELREPOSITORYPARAMETER._serialized_end = 2915
    _REPOSITORYINDEXREQUEST._serialized_start = 2917
    _REPOSITORYINDEXREQUEST._serialized_end = 2981
    _REPOSITORYINDEXRESPONSE._serialized_start = 2984
    _REPOSITORYINDEXRESPONSE._serialized_end = 3148
    _REPOSITORYINDEXRESPONSE_MODELINDEX._serialized_start = 3074
    _REPOSITORYINDEXRESPONSE_MODELINDEX._serialized_end = 3148
    _REPOSITORYMODELLOADREQUEST._serialized_start = 3151
    _REPOSITORYMODELLOADREQUEST._serialized_end = 3387
    _REPOSITORYMODELLOADREQUEST_PARAMETERSENTRY._serialized_start = 3301
    _REPOSITORYMODELLOADREQUEST_PARAMETERSENTRY._serialized_end = 3387
    _REPOSITORYMODELLOADRESPONSE._serialized_start = 3389
    _REPOSITORYMODELLOADRESPONSE._serialized_end = 3418
    _REPOSITORYMODELUNLOADREQUEST._serialized_start = 3421
    _REPOSITORYMODELUNLOADREQUEST._serialized_end = 3661
    _REPOSITORYMODELUNLOADREQUEST_PARAMETERSENTRY._serialized_start = 3301
    _REPOSITORYMODELUNLOADREQUEST_PARAMETERSENTRY._serialized_end = 3387
    _REPOSITORYMODELUNLOADRESPONSE._serialized_start = 3663
    _REPOSITORYMODELUNLOADRESPONSE._serialized_end = 3694
    _GRPCINFERENCESERVICE._serialized_start = 3697
    _GRPCINFERENCESERVICE._serialized_end = 4511
# @@protoc_insertion_point(module_scope)



import json
import logging
import sys

from logging import Formatter, StreamHandler
from pathlib import Path
from typing import Optional, Dict, Union
import logging.config

from .settings import Settings

LoggerName = "mlserver"
LoggerFormat = "%(asctime)s [%(name)s] %(levelname)s - %(message)s"

logger = logging.getLogger(LoggerName)


def get_logger():
    return logger


def apply_logging_file(logging_settings: Union[str, Dict]):
    if isinstance(logging_settings, str) and Path(logging_settings).is_file():
        if "json" in Path(logging_settings).suffix:
            with open(logging_settings) as settings_file:
                config = json.load(settings_file)
            logging.config.dictConfig(config)
        else:
            logging.config.fileConfig(
                fname=logging_settings, disable_existing_loggers=False
            )
    elif isinstance(logging_settings, Dict):
        logging.config.dictConfig(logging_settings)
    else:
        logger.warning("Unable to parse logging_settings.")


def configure_logger(settings: Optional[Settings] = None):
    logger = get_logger()

    # Don't add handler twice
    if not logger.handlers:
        stream_handler = StreamHandler(sys.stdout)
        formatter = Formatter(LoggerFormat)
        stream_handler.setFormatter(formatter)

        logger.addHandler(stream_handler)

    logger.setLevel(logging.INFO)
    if settings and settings.debug:
        logger.setLevel(logging.DEBUG)

    if settings and settings.logging_settings:
        apply_logging_file(settings.logging_settings)

    return logger



from typing import Any, Callable, Optional
from enum import Enum
from functools import wraps

from ..errors import InferenceError
from ..handlers.custom import get_custom_handlers, register_custom_handler
from ..model import MLModel
from ..types import MetadataModelResponse, InferenceRequest, InferenceResponse

from .messages import ModelRequestMessage
from .dispatcher import Dispatcher


class ModelMethods(Enum):
    Predict = "predict"
    Metadata = "metadata"


class ParallelModel(MLModel):
    def __init__(self, model: MLModel, dispatcher: Dispatcher):
        super().__init__(model.settings)
        self._model = model
        self._dispatcher = dispatcher
        self._metadata: Optional[MetadataModelResponse] = None
        self._add_custom_handlers()

    def _add_custom_handlers(self):
        custom_handlers = get_custom_handlers(self._model)
        for handler, method in custom_handlers:
            parallelised = self._parallelise(method)
            register_custom_handler(handler, parallelised)
            setattr(self, method.__name__, parallelised)

    def _parallelise(self, method: Callable):
        @wraps(method)
        async def _inner(*args, **kwargs):
            return await self._send(method.__name__, *args, **kwargs)

        return _inner

    async def metadata(self) -> MetadataModelResponse:
        # NOTE: We cache metadata to avoid expensive worker calls for static
        # data
        if self._metadata is None:
            metadata_response = await self._send(ModelMethods.Metadata.value)
            if not isinstance(metadata_response, MetadataModelResponse):
                raise InferenceError(f"Model '{self.name}' returned no metadata")

            self._metadata = metadata_response

        return self._metadata

    async def predict(self, payload: InferenceRequest) -> InferenceResponse:
        inference_response = await self._send(ModelMethods.Predict.value, payload)
        if not isinstance(inference_response, InferenceResponse):
            raise InferenceError(
                f"Model '{self.name}' returned no predictions after inference"
            )

        return inference_response

    async def _send(self, method_name: str, *args, **kwargs) -> Optional[Any]:
        req_message = ModelRequestMessage(
            model_name=self.name,
            model_version=self.version,
            method_name=method_name,
            method_args=args,
            method_kwargs=kwargs,
        )

        response_message = await self._dispatcher.dispatch_request(req_message)
        return response_message.return_value



import functools

from typing import Any, ClassVar, Dict, Iterable, Optional, Type, Union

from ..types import InferenceRequest, InferenceResponse, RequestInput, ResponseOutput
from ..logging import logger

InputCodecLike = Union[Type["InputCodec"], "InputCodec"]
RequestCodecLike = Union[Type["RequestCodec"], "RequestCodec"]


def deprecated(reason: str):
    def _deprecated(func):
        @functools.wraps(func)
        def _inner(*args, **kwargs):
            logger.warning(f"DEPRECATED! {reason}")
            return func(*args, **kwargs)

        return _inner

    return _deprecated


class InputCodec:
    """
    The InputCodec interface lets you define type conversions of your raw input
    data to / from the Open Inference Protocol.
    Note that this codec applies at the individual input (output) level.

    For request-wide transformations (e.g. dataframes), use the
    ``RequestCodec`` interface instead.
    """

    ContentType: ClassVar[str] = ""
    TypeHint: ClassVar[Type] = type(None)

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        """
        Evaluate whether the codec can encode (decode) the payload.
        """
        return False

    @classmethod
    @deprecated("The encode() method is now deprecated. Use encode_output() instead.")
    def encode(cls, name: str, payload: Any) -> ResponseOutput:
        return cls.encode_output(name, payload)

    @classmethod
    def encode_output(cls, name: str, payload: Any, **kwargs) -> ResponseOutput:
        """
        Encode the given payload into a response output.
        """
        raise NotImplementedError()

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> Any:
        """
        Decode a response output into a high-level Python type.
        """
        raise NotImplementedError()

    @classmethod
    def encode_input(cls, name: str, payload: Any, **kwargs) -> RequestInput:
        """
        Encode the given payload into a ``RequestInput``.
        """
        raise NotImplementedError()

    @classmethod
    @deprecated("The decode() method is now deprecated. Use decode_input() instead.")
    def decode(cls, request_input: RequestInput) -> Any:
        return cls.decode_input(request_input)

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> Any:
        """
        Decode a request input into a high-level Python type.
        """
        raise NotImplementedError()


class RequestCodec:
    """
    The ``RequestCodec`` interface lets you define request-level conversions
    between high-level Python types and the Open Inference Protocol.
    This can be useful where the encoding of your payload encompases multiple
    input heads (e.g. dataframes, where each column can be thought as a
    separate input head).

    For individual input-level encoding / decoding, use the ``InputCodec``
    interface instead.
    """

    ContentType: ClassVar[str] = ""
    TypeHint: ClassVar[Type] = type(None)

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        """
        Evaluate whether the codec can encode (decode) the payload.
        """
        return False

    @classmethod
    @deprecated("The encode() method is now deprecated. Use encode_response() instead.")
    def encode(
        cls, model_name: str, payload: Any, model_version: Optional[str] = None
    ) -> InferenceResponse:
        return cls.encode_response(model_name, payload, model_version)

    @classmethod
    def encode_response(
        cls,
        model_name: str,
        payload: Any,
        model_version: Optional[str] = None,
        **kwargs,
    ) -> InferenceResponse:
        """
        Encode the given payload into an inference response.
        """
        raise NotImplementedError()

    @classmethod
    def decode_response(cls, response: InferenceResponse) -> Any:
        """
        Decode an inference response into a high-level Python object.
        """
        raise NotImplementedError()

    @classmethod
    def encode_request(cls, payload: Any, **kwargs) -> InferenceRequest:
        """
        Encode the given payload into an inference request.
        """
        raise NotImplementedError()

    @classmethod
    @deprecated("The decode() method is now deprecated. Use decode_request() instead.")
    def decode(cls, request: InferenceRequest) -> Any:
        return cls.decode_request(request)

    @classmethod
    def decode_request(cls, request: InferenceRequest) -> Any:
        """
        Decode an inference request into a high-level Python object.
        """
        raise NotImplementedError()


def _find_codec_by_payload(
    payload: Any, codecs: Iterable[Union[RequestCodec, InputCodec]]
) -> Optional[Union[RequestCodec, InputCodec]]:
    matching_codecs = []
    for codec in codecs:
        if codec.can_encode(payload):
            matching_codecs.append(codec)

    if len(matching_codecs) == 0:
        logger.warning(f"No codec was found for payload with type {type(payload)}.")
        return None

    first_codec = matching_codecs[0]
    if len(matching_codecs) > 1:
        logger.warning(
            f"Multiple codecs ({len(matching_codecs)} were found for a "
            f"payload with type {type(payload)}. "
            f"The first one found will be used ({first_codec.ContentType})."
        )

    return first_codec


def _find_codec_by_type_hint(
    type_hint: Type, codecs: Iterable[Union[RequestCodec, InputCodec]]
) -> Optional[Union[RequestCodec, InputCodec]]:
    matching_codecs = []
    for codec in codecs:
        if codec.TypeHint == type_hint:
            matching_codecs.append(codec)

    if len(matching_codecs) == 0:
        logger.warning(f"No codec was found for type hint {type_hint}.")
        return None

    first_codec = matching_codecs[0]
    if len(matching_codecs) > 1:
        logger.warning(
            f"Multiple codecs ({len(matching_codecs)} were found for a "
            f"type hint {type_hint}. "
            f"The first one found will be used ({first_codec.ContentType})."
        )

    return first_codec


class _CodecRegistry:
    """
    CodecRegistry is a "fancy" dictionary to register and find codecs.
    This class has a singleton instance exposed at the module leve, which
    should be used preferably.
    """

    def __init__(
        self,
        input_codecs: Dict[str, InputCodecLike] = {},
        request_codecs: Dict[str, RequestCodecLike] = {},
    ):
        self._input_codecs = input_codecs
        self._request_codecs = request_codecs

    def register_input_codec(self, content_type: str, codec: InputCodecLike):
        # TODO: Raise error if codec exists?
        self._input_codecs[content_type] = codec

    def find_input_codec(
        self,
        content_type: Optional[str] = None,
        payload: Optional[Any] = None,
        type_hint: Optional[Type] = None,
    ) -> Optional[InputCodecLike]:
        if content_type:
            return self._input_codecs.get(content_type)
        elif payload:
            return self.find_input_codec_by_payload(payload)
        elif type_hint:
            return self.find_input_codec_by_type_hint(type_hint)

        return None

    def find_input_codec_by_payload(self, payload: Any) -> Optional[InputCodecLike]:
        return _find_codec_by_payload(
            payload, self._input_codecs.values()  # type: ignore
        )

    def find_input_codec_by_type_hint(
        self, type_hint: Type
    ) -> Optional[InputCodecLike]:
        return _find_codec_by_type_hint(
            type_hint, self._input_codecs.values()  # type: ignore
        )

    def register_request_codec(self, content_type: str, codec: RequestCodecLike):
        # TODO: Raise error if codec exists?
        self._request_codecs[content_type] = codec

    def find_request_codec(
        self,
        content_type: Optional[str] = None,
        payload: Optional[Any] = None,
        type_hint: Optional[Type] = None,
    ) -> Optional[RequestCodecLike]:
        if content_type:
            return self._request_codecs.get(content_type)
        elif payload:
            return self.find_request_codec_by_payload(payload)
        elif type_hint:
            return self.find_request_codec_by_type_hint(type_hint)

        return None

    def find_request_codec_by_payload(self, payload: Any) -> Optional[RequestCodecLike]:
        return _find_codec_by_payload(
            payload, self._request_codecs.values()  # type: ignore
        )

    def find_request_codec_by_type_hint(
        self, type_hint: Type
    ) -> Optional[RequestCodecLike]:
        return _find_codec_by_type_hint(
            type_hint, self._request_codecs.values()  # type: ignore
        )


_codec_registry = _CodecRegistry()

find_request_codec = _codec_registry.find_request_codec
find_request_codec_by_payload = _codec_registry.find_request_codec_by_payload
find_request_codec_by_type_hint = _codec_registry.find_request_codec_by_type_hint
find_input_codec = _codec_registry.find_input_codec
find_input_codec_by_payload = _codec_registry.find_input_codec_by_payload
find_input_codec_by_type_hint = _codec_registry.find_input_codec_by_type_hint


def register_request_codec(CodecKlass: RequestCodecLike):
    if CodecKlass.ContentType is not None:
        _codec_registry.register_request_codec(CodecKlass.ContentType, CodecKlass)
    return CodecKlass


def register_input_codec(CodecKlass: InputCodecLike):
    if CodecKlass.ContentType is not None:
        _codec_registry.register_input_codec(CodecKlass.ContentType, CodecKlass)
    return CodecKlass



import abc
import os
import glob

from pydantic.error_wrappers import ValidationError
from typing import List

from ..settings import ModelParameters, ModelSettings
from ..errors import ModelNotFound
from ..logging import logger

from .load import load_model_settings

DEFAULT_MODEL_SETTINGS_FILENAME = "model-settings.json"


class ModelRepository(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def list(self) -> List[ModelSettings]:
        raise NotImplementedError

    @abc.abstractmethod
    async def find(self, name: str) -> List[ModelSettings]:
        raise NotImplementedError


class SchemalessModelRepository(ModelRepository):
    """
    Model repository, responsible of the discovery of models which can be
    loaded onto the model registry.
    """

    def __init__(self, root: str):
        self._root = root

    async def list(self) -> List[ModelSettings]:
        all_model_settings = []

        # TODO: Use an async alternative for filesys ops
        if self._root:
            abs_root = os.path.abspath(self._root)
            pattern = os.path.join(abs_root, "**", DEFAULT_MODEL_SETTINGS_FILENAME)
            matches = glob.glob(pattern, recursive=True)

            for model_settings_path in matches:
                try:
                    model_settings = load_model_settings(model_settings_path)
                    all_model_settings.append(model_settings)
                except Exception:
                    logger.exception(
                        f"Failed load model settings at {model_settings_path}."
                    )

        # If there were no matches, try to load model from environment
        if not all_model_settings:
            logger.debug(f"No models were found in repository at {self._root}.")
            try:
                # return default
                model_settings = ModelSettings()
                model_settings.parameters = ModelParameters()
                all_model_settings.append(model_settings)
            except ValidationError:
                logger.debug("No default model found in environment settings.")

        return all_model_settings

    async def find(self, name: str) -> List[ModelSettings]:
        all_settings = await self.list()
        selected = []
        for model_settings in all_settings:
            # TODO: Implement other version policies (e.g. "Last N")
            if model_settings.name == name:
                selected.append(model_settings)

        if len(selected) == 0:
            raise ModelNotFound(name)

        return selected



import grpc

from . import dataplane_pb2 as pb
from .dataplane_pb2_grpc import GRPCInferenceServiceServicer
from .converters import (
    ModelInferRequestConverter,
    ModelInferResponseConverter,
    ServerMetadataResponseConverter,
    ModelMetadataResponseConverter,
    RepositoryIndexRequestConverter,
    RepositoryIndexResponseConverter,
)
from .utils import to_headers, to_metadata, handle_mlserver_error

from ..utils import insert_headers, extract_headers
from ..handlers import DataPlane, ModelRepositoryHandlers


class InferenceServicer(GRPCInferenceServiceServicer):
    def __init__(
        self, data_plane: DataPlane, model_repository_handlers: ModelRepositoryHandlers
    ):
        super().__init__()
        self._data_plane = data_plane
        self._model_repository_handlers = model_repository_handlers

    async def ServerLive(
        self, request: pb.ServerLiveRequest, context
    ) -> pb.ServerLiveResponse:
        is_live = await self._data_plane.live()
        return pb.ServerLiveResponse(live=is_live)

    async def ServerReady(
        self, request: pb.ServerReadyRequest, context
    ) -> pb.ServerReadyResponse:
        is_ready = await self._data_plane.ready()
        return pb.ServerReadyResponse(ready=is_ready)

    async def ModelReady(
        self, request: pb.ModelReadyRequest, context
    ) -> pb.ModelReadyResponse:
        is_model_ready = await self._data_plane.model_ready(
            name=request.name, version=request.version
        )
        return pb.ModelReadyResponse(ready=is_model_ready)

    async def ServerMetadata(
        self, request: pb.ServerMetadataRequest, context
    ) -> pb.ServerMetadataResponse:
        metadata = await self._data_plane.metadata()
        return ServerMetadataResponseConverter.from_types(metadata)

    @handle_mlserver_error
    async def ModelMetadata(
        self, request: pb.ModelMetadataRequest, context
    ) -> pb.ModelMetadataResponse:
        metadata = await self._data_plane.model_metadata(
            name=request.name, version=request.version
        )
        return ModelMetadataResponseConverter.from_types(metadata)

    @handle_mlserver_error
    async def ModelInfer(
        self, request: pb.ModelInferRequest, context: grpc.ServicerContext
    ) -> pb.ModelInferResponse:
        return_raw = False
        if request.raw_input_contents:
            # If the request contains raw input contents, then use the same for
            # the output
            return_raw = True

        payload = ModelInferRequestConverter.to_types(request)

        request_headers = to_headers(context)
        insert_headers(payload, request_headers)

        result = await self._data_plane.infer(
            payload=payload, name=request.model_name, version=request.model_version
        )

        response_headers = extract_headers(result)
        if response_headers:
            response_metadata = to_metadata(response_headers)
            context.set_trailing_metadata(response_metadata)

        response = ModelInferResponseConverter.from_types(result, use_raw=return_raw)
        return response

    async def RepositoryIndex(
        self, request: pb.RepositoryIndexRequest, context
    ) -> pb.RepositoryIndexResponse:
        payload = RepositoryIndexRequestConverter.to_types(request)
        index = await self._model_repository_handlers.index(payload)
        return RepositoryIndexResponseConverter.from_types(index)  # type: ignore

    @handle_mlserver_error
    async def RepositoryModelLoad(
        self, request: pb.RepositoryModelLoadRequest, context
    ) -> pb.RepositoryModelLoadResponse:
        await self._model_repository_handlers.load(request.model_name)
        return pb.RepositoryModelLoadResponse()

    @handle_mlserver_error
    async def RepositoryModelUnload(
        self, request: pb.RepositoryModelUnloadRequest, context
    ) -> pb.RepositoryModelUnloadResponse:
        await self._model_repository_handlers.unload(request.model_name)
        return pb.RepositoryModelUnloadResponse()



from typing import Callable
from fastapi import FastAPI
from fastapi.responses import Response as FastAPIResponse
from fastapi.routing import APIRoute as FastAPIRoute
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from starlette_exporter import PrometheusMiddleware

from .endpoints import Endpoints, ModelRepositoryEndpoints
from .requests import Request
from .responses import Response
from .errors import _EXCEPTION_HANDLERS

from ..settings import Settings
from ..handlers import DataPlane, ModelRepositoryHandlers


class APIRoute(FastAPIRoute):
    """
    Custom route to use our own Request handler.
    """

    def __init__(
        self,
        *args,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        **kwargs
    ):
        super().__init__(
            *args,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_none=response_model_exclude_none,
            **kwargs
        )

    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> FastAPIResponse:
            request = Request(request.scope, request.receive)
            return await original_route_handler(request)

        return custom_route_handler


def create_app(
    settings: Settings,
    data_plane: DataPlane,
    model_repository_handlers: ModelRepositoryHandlers,
) -> FastAPI:
    endpoints = Endpoints(data_plane)
    model_repository_endpoints = ModelRepositoryEndpoints(model_repository_handlers)

    routes = [
        # Model ready
        APIRoute(
            "/v2/models/{model_name}/ready",
            endpoints.model_ready,
        ),
        APIRoute(
            "/v2/models/{model_name}/versions/{model_version}/ready",
            endpoints.model_ready,
        ),
        # Model infer
        APIRoute(
            "/v2/models/{model_name}/infer",
            endpoints.infer,
            methods=["POST"],
        ),
        APIRoute(
            "/v2/models/{model_name}/versions/{model_version}/infer",
            endpoints.infer,
            methods=["POST"],
        ),
        # Model metadata
        APIRoute(
            "/v2/models/{model_name}",
            endpoints.model_metadata,
        ),
        APIRoute(
            "/v2/models/{model_name}/versions/{model_version}",
            endpoints.model_metadata,
        ),
        # Liveness and readiness
        APIRoute("/v2/health/live", endpoints.live),
        APIRoute("/v2/health/ready", endpoints.ready),
        # Server metadata
        APIRoute(
            "/v2",
            endpoints.metadata,
        ),
    ]

    routes += [
        # Model Repository API
        APIRoute(
            "/v2/repository/index",
            model_repository_endpoints.index,
            methods=["POST"],
        ),
        APIRoute(
            "/v2/repository/models/{model_name}/load",
            model_repository_endpoints.load,
            methods=["POST"],
        ),
        APIRoute(
            "/v2/repository/models/{model_name}/unload",
            model_repository_endpoints.unload,
            methods=["POST"],
        ),
    ]

    app = FastAPI(
        debug=settings.debug,
        routes=routes,  # type: ignore
        default_response_class=Response,
        exception_handlers=_EXCEPTION_HANDLERS,  # type: ignore
    )
    app.router.route_class = APIRoute
    app.add_middleware(GZipMiddleware)
    if settings.cors_settings is not None:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=settings.cors_settings.allow_origins,
            allow_origin_regex=settings.cors_settings.allow_origin_regex,
            allow_credentials=settings.cors_settings.allow_credentials,
            allow_methods=settings.cors_settings.allow_methods,
            allow_headers=settings.cors_settings.allow_headers,
            max_age=settings.cors_settings.max_age,
        )

    if settings.metrics_endpoint:
        app.add_middleware(
            PrometheusMiddleware,
            app_name="mlserver",
            prefix=settings.metrics_rest_server_prefix,
            # TODO: Should we also exclude model's health endpoints?
            skip_paths=[
                settings.metrics_endpoint,
                "/v2/health/live",
                "/v2/health/ready",
            ],
        )

    return app



import base64
import binascii

from typing import Any, List, Union
from functools import partial

from ..types import RequestInput, ResponseOutput
from .lists import is_list_of
from .base import InputCodec, register_input_codec
from .lists import as_list, ListElement

_Base64StrCodec = "ascii"


def _ensure_bytes(elem: ListElement) -> bytes:
    if isinstance(elem, str):
        return elem.encode(_Base64StrCodec)

    return elem


def _encode_base64(elem: ListElement, use_bytes: bool) -> Union[bytes, str]:
    as_bytes = _ensure_bytes(elem)
    b64_encoded = base64.b64encode(as_bytes)
    if use_bytes:
        return b64_encoded

    return b64_encoded.decode(_Base64StrCodec)


def _decode_base64(elem: ListElement) -> bytes:
    as_bytes = _ensure_bytes(elem)

    # Check that the input is valid base64.
    # Otherwise, convert into base64.
    try:
        return base64.b64decode(as_bytes, validate=True)
    except binascii.Error:
        return as_bytes


@register_input_codec
class Base64Codec(InputCodec):
    """
    Codec that convers to / from a base64 input.
    """

    ContentType = "base64"
    TypeHint = List[bytes]

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return is_list_of(payload, bytes)

    @classmethod
    def encode_output(
        cls, name: str, payload: List[bytes], use_bytes: bool = True, **kwargs
    ) -> ResponseOutput:
        # Assume that payload is already in b64, so we only need to pack it
        packed = map(partial(_encode_base64, use_bytes=use_bytes), payload)
        shape = [len(payload), 1]
        return ResponseOutput(
            name=name,
            datatype="BYTES",
            shape=shape,
            data=list(packed),
        )

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> List[bytes]:
        packed = response_output.data.__root__
        return list(map(_decode_base64, as_list(packed)))

    @classmethod
    def encode_input(
        cls, name: str, payload: List[bytes], use_bytes: bool = True, **kwargs
    ) -> RequestInput:
        # Assume that payload is already in b64, so we only need to pack it
        output = cls.encode_output(name, payload, use_bytes)
        return RequestInput(
            name=output.name,
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
        )

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> List[bytes]:
        packed = request_input.data.__root__

        return list(map(_decode_base64, as_list(packed)))



from .numpy import NumpyCodec, NumpyRequestCodec
from .pandas import PandasCodec
from .string import StringCodec
from .base64 import Base64Codec
from .datetime import DatetimeCodec
from .errors import CodecError
from .decorator import decode_args
from .base import (
    InputCodec,
    RequestCodec,
    register_input_codec,
    register_request_codec,
    InputCodecLike,
    RequestCodecLike,
)
from .utils import (
    DecodedParameterName,
    has_decoded,
    get_decoded,
    get_decoded_or_raw,
    encode_inference_response,
    encode_response_output,
    decode_request_input,
    decode_inference_request,
)

__all__ = [
    "CodecError",
    "NumpyCodec",
    "NumpyRequestCodec",
    "StringCodec",
    "StringRequestCodec",
    "Base64Codec",
    "DatetimeCodec",
    "PandasCodec",
    "InputCodec",
    "InputCodecLike",
    "RequestCodec",
    "RequestCodecLike",
    "DecodedParameterName",
    "register_input_codec",
    "register_request_codec",
    "has_decoded",
    "get_decoded",
    "get_decoded_or_raw",
    "encode_inference_response",
    "encode_response_output",
    "decode_request_input",
    "decode_inference_request",
    "decode_args",
]



from pydantic import BaseModel as _BaseModel


class BaseModel(_BaseModel):
    """
    Override Pydantic's BaseModel class to ensure all payloads exclude unset
    fields by default.

    From:
        https://github.com/pydantic/pydantic/issues/1387#issuecomment-612901525
    """

    def dict(self, exclude_unset=True, exclude_none=True, **kwargs):
        return super().dict(
            exclude_unset=exclude_unset, exclude_none=exclude_none, **kwargs
        )

    def json(self, exclude_unset=True, exclude_none=True, **kwargs):
        return super().json(
            exclude_unset=exclude_unset, exclude_none=exclude_none, **kwargs
        )



from typing import Awaitable, Callable, Tuple
from functools import partial
from timeit import default_timer

from grpc.aio import ServerInterceptor, ServicerContext
from grpc import HandlerCallDetails, RpcMethodHandler, RpcError, StatusCode
from py_grpc_prometheus.prometheus_server_interceptor import (
    grpc_utils,
    PromServerInterceptor as _PromServerInterceptor,
)

from .logging import logger


class LoggingInterceptor(ServerInterceptor):
    def _get_log_message(self, handler_call_details: HandlerCallDetails) -> str:
        return handler_call_details.method

    async def intercept_service(
        self,
        continuation: Callable[[HandlerCallDetails], Awaitable[RpcMethodHandler]],
        handler_call_details: HandlerCallDetails,
    ) -> RpcMethodHandler:
        logger.info(self._get_log_message(handler_call_details))
        handler = await continuation(handler_call_details)
        return handler


class PromServerInterceptor(ServerInterceptor):
    """
    Simple wrapper around `py_grpc_prometheus` to support `grpc.aio`.

    TODO: Open PR to add support upstream for AsyncIO.
    """

    def __init__(self, *args, **kwargs):
        self._interceptor = _PromServerInterceptor(*args, **kwargs)
        # We need a status code mapping to ensure we can convert from an int:
        # https://groups.google.com/g/grpc-io/c/EdIXjMEaOyw/m/d3DeqmrJAAAJ
        self._status_codes = {code.value[0]: code for code in StatusCode}

    async def intercept_service(
        self,
        continuation: Callable[[HandlerCallDetails], Awaitable[RpcMethodHandler]],
        handler_call_details: HandlerCallDetails,
    ) -> RpcMethodHandler:
        method_call = grpc_utils.split_method_call(handler_call_details)
        handler = await continuation(handler_call_details)

        metrics_wrapper = partial(self._metrics_wrapper, method_call)
        return self._interceptor._wrap_rpc_behavior(handler, metrics_wrapper)

    def _compute_status_code(self, servicer_context: ServicerContext) -> StatusCode:
        """
        This method is mostly copied from `py-grpc-prometheus`, with a couple
        minor changes to avoid using private APIs from ServicerContext which
        don't exist anymore in `grpc.aio`.
        To see the original implementation, please check:

        https://github.com/lchenn/py-grpc-prometheus/blob/eb9dee1f0a4e57cef220193ee48021dc9a9f3d82/py_grpc_prometheus/prometheus_server_interceptor.py#L127-L134
        """
        # Backwards compatibility for non-aio.
        # TODO: It's not clear yet how to check whether the context has been
        # cancelled with aio.
        if hasattr(servicer_context, "_state"):
            if servicer_context._state.client == "cancelled":
                return StatusCode.CANCELLED

        if not hasattr(servicer_context, "code"):
            return StatusCode.OK

        code = servicer_context.code()
        if code is None:
            return StatusCode.OK

        # NOTE: With gRPC AIO, the `code` can be a plain integer that needs to
        # be converted to an actual `StatusCode` entry
        if isinstance(code, int):
            if code not in self._status_codes:
                return StatusCode.UNKNOWN

            return self._status_codes[code]

        return code

    def _metrics_wrapper(
        self,
        method_call: Tuple[str, str, str],
        old_handler: RpcMethodHandler,
        request_streaming: bool,
        response_streaming: bool,
    ):
        """
        Port of `py-grpc-prometheus` metrics_wrapper method to work with gRPC's
        AsyncIO support.
        To see the original implementation, please check:

        https://github.com/lchenn/py-grpc-prometheus/blob/eb9dee1f0a4e57cef220193ee48021dc9a9f3d82/py_grpc_prometheus/prometheus_server_interceptor.py#L46-L120
        """
        grpc_service_name, grpc_method_name, _ = method_call

        async def _new_handler(request_or_iterator, servicer_context: ServicerContext):
            response_or_iterator = None
            try:
                start = default_timer()
                grpc_type = grpc_utils.get_method_type(
                    request_streaming, response_streaming
                )
                try:
                    if request_streaming:
                        request_or_iterator = grpc_utils.wrap_iterator_inc_counter(
                            request_or_iterator,
                            self._interceptor._metrics[
                                "grpc_server_stream_msg_received"
                            ],
                            grpc_type,
                            grpc_service_name,
                            grpc_method_name,
                        )
                    else:
                        self._interceptor._metrics[
                            "grpc_server_started_counter"
                        ].labels(
                            grpc_type=grpc_type,
                            grpc_service=grpc_service_name,
                            grpc_method=grpc_method_name,
                        ).inc()

                    # Invoke the original rpc behavior.
                    # NOTE: This is the main change required with respect to
                    # the original implementation in `py-grpc-prometheus`.
                    response_or_iterator = await old_handler(
                        request_or_iterator, servicer_context
                    )

                    if response_streaming:
                        sent_metric = self._interceptor._metrics[
                            "grpc_server_stream_msg_sent"
                        ]
                        response_or_iterator = grpc_utils.wrap_iterator_inc_counter(
                            response_or_iterator,
                            sent_metric,
                            grpc_type,
                            grpc_service_name,
                            grpc_method_name,
                        )

                    else:
                        self._interceptor.increase_grpc_server_handled_total_counter(
                            grpc_type,
                            grpc_service_name,
                            grpc_method_name,
                            self._compute_status_code(servicer_context).name,
                        )
                    return response_or_iterator
                except RpcError as e:
                    self._interceptor.increase_grpc_server_handled_total_counter(
                        grpc_type,
                        grpc_service_name,
                        grpc_method_name,
                        self._interceptor._compute_error_code(e).name,
                    )
                    raise e

                finally:
                    if not response_streaming:
                        if self._interceptor._legacy:
                            self._interceptor._metrics[
                                "legacy_grpc_server_handled_latency_seconds"
                            ].labels(
                                grpc_type=grpc_type,
                                grpc_service=grpc_service_name,
                                grpc_method=grpc_method_name,
                            ).observe(
                                max(default_timer() - start, 0)
                            )
                        elif self._interceptor._enable_handling_time_histogram:
                            self._interceptor._metrics[
                                "grpc_server_handled_histogram"
                            ].labels(
                                grpc_type=grpc_type,
                                grpc_service=grpc_service_name,
                                grpc_method=grpc_method_name,
                            ).observe(
                                max(default_timer() - start, 0)
                            )
            except Exception as e:  # pylint: disable=broad-except
                # Allow user to skip the exceptions in order to maintain
                # the basic functionality in the server
                # The logging function in exception can be toggled with log_exceptions
                # in order to suppress the noise in logging
                if self._interceptor._skip_exceptions:
                    if self._interceptor._log_exceptions:
                        logger.error(e)
                    if response_or_iterator is None:
                        return response_or_iterator
                    return old_handler(request_or_iterator, servicer_context)
                raise e

        return _new_handler



import asyncio
import os
import sys
import tarfile
import glob

from typing import List
from functools import cached_property

from .logging import logger


def _extract_env(tarball_path: str, env_path: str) -> None:
    logger.info(f"Extracting environment tarball from {tarball_path}...")
    with tarfile.open(tarball_path, "r") as tarball:
        tarball.extractall(path=env_path)


class Environment:
    """
    Custom Python environment.
    The class can be used as a context manager to enable / disable the custom
    environment.
    """

    def __init__(self, env_path: str):
        self._env_path = env_path

    @classmethod
    async def from_tarball(cls, tarball_path: str, env_path: str) -> "Environment":
        """
        Instantiate an Environment object from an environment tarball.
        """
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, _extract_env, tarball_path, env_path)

        return cls(env_path)

    @cached_property
    def _sys_path(self) -> List[str]:
        """
        Extra paths that will be added to `sys.path` (i.e. `PYTHONPATH`) to
        expose the custom environment.
        These paths are mainly used on the `__enter__` method of the context
        manager below.
        """
        if not self._lib_path:
            return []

        # This list of paths is mainly built from PEP370 (and also from how
        # PYTHONPATH is defined by default).
        # The main route is probably just `.../site-packages`, but we add the
        # rest to be safe.
        return [
            f"{self._lib_path}.zip",
            self._lib_path,
            os.path.join(self._lib_path, "lib-dynload"),
            os.path.join(self._lib_path, "site-packages"),
        ]

    @cached_property
    def _bin_path(self) -> str:
        """
        Path to the `../bin/` folder in our custom environment.
        """
        return os.path.join(self._env_path, "bin")

    @cached_property
    def _lib_path(self) -> str:
        """
        Base environment path (i.e. user data directory - as defined by
        PEP370).
        """
        pattern = os.path.join(self._env_path, "lib", "python*")
        matches = glob.glob(pattern)

        for match in matches:
            if os.path.isdir(match):
                return match

        return ""

    def __enter__(self):
        self._prev_sys_path = sys.path
        self._prev_bin_path = os.environ["PATH"]

        sys.path = [*self._sys_path, *self._prev_sys_path]
        os.environ["PATH"] = os.pathsep.join([self._bin_path, self._prev_bin_path])

        return self

    def __exit__(self, *exc_details) -> None:
        sys.path = self._prev_sys_path
        os.environ["PATH"] = self._prev_bin_path



import glob
import os
import asyncio

from aiofiles.os import remove
from prometheus_client import (
    REGISTRY,
    CollectorRegistry,
    generate_latest,
    CONTENT_TYPE_LATEST,
    values,
)
from prometheus_client.multiprocess import MultiProcessCollector, mark_process_dead
from fastapi import Request, Response, status

from ..settings import Settings

PROMETHEUS_MULTIPROC_DIR = "PROMETHEUS_MULTIPROC_DIR"


def configure_metrics(settings: Settings):
    if not settings.parallel_workers:
        return

    # Re-set Prometheus' Value class to use the multiproc version.
    # Note that this workaround depends on initialising all metrics in a
    # lazy manner (i.e. not as global values)
    # https://github.com/prometheus/client_python/blob/781e3e1851d80a53732bb8102d5754cf9d68b3c1/prometheus_client/values.py#L126-L134
    os.environ[PROMETHEUS_MULTIPROC_DIR] = settings.metrics_dir
    values.ValueClass = values.get_value_class()


async def stop_metrics(settings: Settings, pid: int):
    if not settings.parallel_workers:
        return

    mark_process_dead(pid)
    pattern = os.path.join(settings.metrics_dir, f"*_{pid}.db")
    matching_files = glob.glob(pattern)
    await asyncio.gather(*[remove(f) for f in matching_files])


class PrometheusEndpoint:
    def __init__(self, settings: Settings):
        self._settings = settings
        configure_metrics(self._settings)

    @property
    def _registry(self) -> CollectorRegistry:
        if not self._settings.parallel_workers:
            return REGISTRY

        registry = CollectorRegistry()
        MultiProcessCollector(registry)

        return registry

    def handle_metrics(self, req: Request) -> Response:
        headers = {"Content-Type": CONTENT_TYPE_LATEST}
        return Response(
            generate_latest(self._registry),
            status_code=status.HTTP_200_OK,
            headers=headers,
        )



import inspect

from typing import Tuple, Any, Callable, List
from pydantic import BaseModel

from ..model import MLModel

_CustomHandlerAttr = "__custom_handler__"
_HandlerMethod = Callable[..., Any]


class CustomHandler(BaseModel):
    rest_path: str
    rest_method: str = "POST"


def custom_handler(rest_path: str, rest_method: str = "POST"):
    def _wraps(f):
        handler = CustomHandler(rest_path=rest_path, rest_method=rest_method)
        register_custom_handler(handler, f)
        return f

    return _wraps


def register_custom_handler(handler: CustomHandler, method: _HandlerMethod):
    setattr(method, _CustomHandlerAttr, handler)


def get_custom_handlers(model: MLModel) -> List[Tuple[CustomHandler, _HandlerMethod]]:
    handlers = []
    members = inspect.getmembers(model)

    for _, member in members:
        if not hasattr(member, _CustomHandlerAttr):
            continue

        custom_handler = getattr(member, _CustomHandlerAttr)
        handlers.append((custom_handler, member))

    return handlers



from fastapi import status

from ..utils import get_import_path
from ..errors import MLServerError


class WorkerError(MLServerError):
    """
    Class used to wrap exceptions raised from the workers.

    All stacktrace details will be hidden, and the original class won't be
    returned. This is to avoid issues with custom exceptions, like:

        https://github.com/SeldonIO/MLServer/issues/881
    """

    def __init__(self, exc: BaseException):
        msg = str(exc)
        if isinstance(exc, BaseException):
            import_path = get_import_path(exc.__class__)
            msg = f"{import_path}: {exc}"

        status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
        if isinstance(exc, MLServerError):
            status_code = exc.status_code

        super().__init__(msg, status_code)



from dataclasses import dataclass
from functools import wraps
import os
from random import random
import tritonclient.http.aio as httpclient

import asyncio
import aiofiles
import logging
import click
import json
import orjson

from time import perf_counter as timer
from typing import Dict, List, Optional, Tuple

from pydantic.error_wrappers import ValidationError

from mlserver.batching.requests import BatchedRequests
from mlserver.types import InferenceRequest, InferenceResponse, Parameters
from mlserver.codecs import NumpyCodec
from mlserver.logging import get_logger


from mlserver.utils import generate_uuid
from mlserver.batching.requests import _merge_parameters


CHOICES_TRANSPORT = ["rest", "grpc"]
FINALIZER_REPORT_FREQUENCY = 100
RETRY_SLEEP_TIME = 1

logger = get_logger()


# Monkey patching is required for error responses coming from MLServer.
# MLServer will contain `detail` field instead of `error` one like Triton.
async def _get_error(response):
    """
    Returns the InferenceServerException object if response
    indicates the error. If no error then return None
    """
    if response.status != 200:
        result = await response.read()
        error_response = json.loads(result) if len(result) else {"error": ""}
        try:
            return httpclient.InferenceServerException(msg=error_response["error"])
        except KeyError:
            return httpclient.InferenceServerException(
                msg=json.dumps(error_response["detail"])
            )
    else:
        return None


httpclient._get_error = _get_error
del _get_error


def click_async(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@dataclass
class BatchInputItem:
    index: int
    item: bytes


@dataclass
class BatchOutputItem:
    index: int
    item: bytes


def setup_logging(debug: bool):
    if debug:
        logger.setLevel(logging.DEBUG)


def get_headers(request_id: str, headers: Dict[str, str]) -> Dict[str, str]:
    headers = {"content-type": "application/json", **headers}
    if request_id:
        headers["seldon-puid"] = request_id
        headers["x-request-id"] = request_id
    return headers


@dataclass
class TritonRequest:
    id: str
    inputs: List[httpclient.InferInput]
    outputs: Optional[List[httpclient.InferRequestedOutput]]

    @classmethod
    def from_inference_request(
        cls, inference_request: InferenceRequest, binary_data: bool
    ) -> "TritonRequest":
        inputs = []

        for request_input in inference_request.inputs or []:
            new_input = httpclient.InferInput(
                request_input.name, request_input.shape, request_input.datatype
            )
            new_input.set_data_from_numpy(
                NumpyCodec.decode_input(request_input),
                binary_data=binary_data,
            )
            inputs.append(new_input)

        outputs = []
        for request_output in inference_request.outputs or []:
            new_output = httpclient.InferRequestedOutput(
                request_output.name, binary_data=binary_data
            )
            outputs.append(new_output)

        if inference_request.id is None:
            request_id = ""
        else:
            request_id = inference_request.id
        return TritonRequest(request_id, inputs, outputs)


def infer_result_to_infer_response(item: httpclient.InferResult) -> InferenceResponse:
    infer_response = item.get_response()

    outputs = []
    for response_output in infer_response["outputs"]:
        name = response_output["name"]
        output = NumpyCodec.encode_output(name, item.as_numpy(name))

        # Drop "binary_data_size" from `paramaters` as we decoded the output
        parameters = response_output.get("parameters")
        if parameters is not None:
            # Create copy in case this code will ever be used in another context as
            # we are modifying the original parameters by removing unwanted key
            parameters = parameters.copy()
            if "binary_data_size" in parameters:
                del parameters["binary_data_size"]
            output.parameters = Parameters(**parameters)

        outputs.append(output)

    inference_response = InferenceResponse(
        model_name=infer_response["model_name"],
        model_version=infer_response.get("model_version", None),
        id=infer_response.get("id", None),
        parameters=infer_response.get("parameters", None),
        outputs=outputs,
    )

    return inference_response


def _preprocess_headers(headers: List[str]) -> Dict[str, str]:
    output = {}
    for item in headers:
        try:
            key, val = [x.strip() for x in item.split(":")]
            output[key] = val
        except ValueError:
            logger.error(f"Cannot process '{item}' as valid header. Ignoring.")
    return output


def _verify_write_access(fpath: str):
    return os.access(os.path.dirname(os.path.abspath(fpath)), os.W_OK)


def _serialize_validation_error(index: int, error: ValidationError) -> BatchOutputItem:
    msg = {
        "parameters": {"batch_index": index},
        "error": {"status": "preprocessing error", "msg": error.errors()},
    }
    return BatchOutputItem(item=json.dumps(msg).encode(), index=index)


def _serialize_inference_error(index: int, error: Exception) -> BatchOutputItem:
    msg = {
        "parameters": {"batch_index": index},
        "error": {"status": "preprocessing error", "msg": str(error)},
    }
    return BatchOutputItem(item=json.dumps(msg).encode(), index=index)


def preprocess_items(
    items: List[BatchInputItem], binary_data: bool
) -> Tuple[TritonRequest, BatchedRequests, List[BatchOutputItem], Dict[str, int]]:
    item_indices = {}
    inference_requests = {}
    invalid_inputs = []
    for item in items:
        try:
            inference_request = InferenceRequest.parse_obj(orjson.loads(item.item))
            # try to use `id` from the input file to identify each request in the batch
            if inference_request.id is None:
                inference_request.id = generate_uuid()
            inference_requests[inference_request.id] = inference_request
            item_indices[inference_request.id] = item.index
        except ValidationError as e:
            logger.error(
                f"preprocessing error: batch_index={item.index}, error={repr(e)}"
            )
            invalid_inputs.append(_serialize_validation_error(item.index, e))
    batched = BatchedRequests(inference_requests)

    # Set `id` for batched requests - if only single request use its own id
    if len(inference_requests) == 1:
        batched.merged_request.id = inference_request.id
    else:
        batched.merged_request.id = generate_uuid()
    return (
        TritonRequest.from_inference_request(batched.merged_request, binary_data),
        batched,
        invalid_inputs,
        item_indices,
    )


def postprocess_items(
    infer_result: httpclient.InferResult,
    batched: BatchedRequests,
    item_indices: Dict[str, int],
) -> List[BatchOutputItem]:
    full_inference_response = infer_result_to_infer_response(infer_result)
    output_items = []
    for item_id, inference_response in batched.split_response(
        full_inference_response
    ).items():
        # Add `id` used for batched requests to Parameters under `inference_id` key
        new_params = {
            "batch_index": item_indices[item_id],
            "inference_id": full_inference_response.id,
        }
        inference_response.parameters = Parameters(
            **_merge_parameters(new_params, inference_response)
        )

        output_items.append(
            BatchOutputItem(
                index=item_indices[item_id], item=inference_response.json().encode()
            )
        )
    return output_items


async def send_requests(
    model_name: str,
    triton_client: httpclient.InferenceServerClient,
    triton_request: TritonRequest,
    headers: Dict[str, str],
    retries: int,
    worker_id: int,
) -> httpclient.InferResult:
    for i in range(retries):
        try:
            return await triton_client.infer(
                model_name,
                triton_request.inputs,
                outputs=triton_request.outputs,
                request_id=triton_request.id,
                headers=get_headers(triton_request.id, headers),
            )
        except Exception as e:
            logger.error(
                f"consumer {worker_id}: retries {i+1}/{retries}, exception {e}"
            )
            if i + 1 == retries:
                raise
            await asyncio.sleep(RETRY_SLEEP_TIME)


async def process_items(
    items: List[BatchInputItem],
    model_name: str,
    worker_id: int,
    retries: int,
    triton_client: httpclient.InferenceServerClient,
    headers: Dict[str, str],
    binary_data: bool,
) -> List[BatchOutputItem]:
    try:
        triton_request, batched, invalid_inputs, inference_indices = preprocess_items(
            items, binary_data
        )
        if len(invalid_inputs) == len(items):
            return invalid_inputs
    except Exception as e:
        logger.error(f"consumer {worker_id}: failed to preprocess items: {e}")
        raise

    try:
        logger.debug(f"consumer {worker_id}: sending request")
        infer_result = await send_requests(
            model_name,
            triton_client,
            triton_request,
            headers,
            retries,
            worker_id,
        )
        logger.debug(f"consumer {worker_id}: received response")
    except Exception as e:
        logger.error(f"consumer {worker_id}: failed to process task: {e}")
        failed_results = [
            _serialize_inference_error(index, e) for index in inference_indices.values()
        ]
        return failed_results + invalid_inputs

    try:
        output_items = postprocess_items(infer_result, batched, inference_indices)
    except Exception as e:
        logger.error(f"consumer {worker_id}: failed to postprocess item: {e}")

    return output_items + invalid_inputs


async def produce(
    queue: "asyncio.Queue[List[BatchInputItem]]", fname: str, batch_size: int
):
    async with aiofiles.open(fname, "rb") as f:
        index = 0
        batch = []
        async for line in f:
            batch.append(BatchInputItem(index=index, item=line))
            if len(batch) == batch_size:
                await queue.put(batch)
                batch = []
            index += 1
        if batch:
            await queue.put(batch)


async def consume(
    model_name: str,
    worker_id: int,
    retries: int,
    batch_interval: float,
    batch_jitter: float,
    triton_client: httpclient.InferenceServerClient,
    headers: Dict[str, str],
    binary_data: bool,
    extra_verbose: bool,
    queue_in: "asyncio.Queue[List[BatchInputItem]]",
    queue_out: "asyncio.Queue[List[BatchOutputItem]]",
):
    while True:
        input_items = await queue_in.get()
        try:
            start_time = timer()
            output_items = await process_items(
                input_items,
                model_name,
                worker_id,
                retries,
                triton_client,
                headers,
                binary_data,
            )
            if batch_interval > 0 or batch_jitter > 0:
                total_sleep_time = batch_interval + random() * batch_jitter
                remaining_sleep_time = total_sleep_time - (timer() - start_time)
                logger.debug(
                    f"consume {worker_id}: sleeping for {remaining_sleep_time:.3f}"
                )
                await asyncio.sleep(remaining_sleep_time)
        except Exception:
            if extra_verbose:
                logger.error(
                    f"consumer {worker_id}: failed to process item {input_items}"
                )
            queue_in.task_done()
            continue

        await queue_out.put(output_items)
        queue_in.task_done()


async def finalize(queue: "asyncio.Queue[List[BatchOutputItem]]", fname: str) -> int:
    # TODO: Test if output directory is writtable, sysexit otherwise.
    counter = 0
    try:
        async with aiofiles.open(fname, "wb") as f:
            while True:
                batch_output = await queue.get()
                for item in batch_output:
                    try:
                        await f.write(item.item)
                        await f.write(b"\n")
                    except Exception as e:
                        logger.error(f"Failed to finalize task: {repr(e)}")
                    counter += 1
                    if counter % FINALIZER_REPORT_FREQUENCY == 0:
                        logger.info(f"Finalizer: processed instances: {counter}")
                queue.task_done()
    except asyncio.CancelledError:
        return counter


async def process_batch(
    model_name: str,
    url: str,
    workers: int,
    retries: int,
    batch_size: int,
    input_data_path: str,
    output_data_path: str,
    binary_data: bool,
    transport: str,
    request_headers: List[str],
    timeout: float,
    batch_interval: float,
    batch_jitter: float,
    use_ssl: bool,
    insecure: bool,
    verbose: bool,
    extra_verbose: bool,
):
    """
    Process batch requests against V2-compatible inference server.

    Currently, this is meant to be an entrypoint for the `mlserver infer ...` CLI.
    Note: this feature is currently experimental and API is subject to change.
    """
    start_time = timer()

    if transport == "grpc":
        raise click.BadParameter("The 'grpc' transport is not yet supported.")

    if extra_verbose:
        verbose = True
        logger.info("Running in extra verbose mode.")
    elif verbose:
        logger.info("Running in verbose mode.")

    headers = _preprocess_headers(request_headers)

    setup_logging(debug=verbose)
    logger.info(f"server url: {url}")
    logger.info(f"model name: {model_name}")
    logger.info(f"request headers: {headers}")
    logger.info(f"input file path: {input_data_path}")
    logger.info(f"output file path: {output_data_path}")
    logger.info(f"workers: {workers}")
    logger.info(f"retries: {retries}")
    logger.info(f"batch interval: {batch_interval}")
    logger.info(f"batch jitter: {batch_jitter}")
    logger.info(f"connection timeout: {timeout}")
    logger.info(f"micro-batch size: {batch_size}")

    if not _verify_write_access(output_data_path):
        raise click.BadParameter(
            f"Provided output file path '{output_data_path}' is not writable."
        )

    if insecure:
        ssl_context = False
    else:
        ssl_context = None

    triton_client = httpclient.InferenceServerClient(
        url=url,
        verbose=extra_verbose,
        conn_limit=workers,
        conn_timeout=timeout,
        ssl=use_ssl,
        ssl_context=ssl_context,
    )

    queue_in: asyncio.Queue[List[BatchInputItem]] = asyncio.Queue(2 * workers)
    queue_out: asyncio.Queue[List[BatchOutputItem]] = asyncio.Queue(2 * workers)

    consumers = []
    for worker_id in range(workers):
        consumer = asyncio.create_task(
            consume(
                model_name,
                worker_id,
                retries,
                batch_interval,
                batch_jitter,
                triton_client,
                headers,
                binary_data,
                extra_verbose,
                queue_in,
                queue_out,
            )
        )
        consumers.append(consumer)

    finalizer = asyncio.create_task(finalize(queue_out, output_data_path))

    await produce(queue_in, input_data_path, batch_size)
    await queue_in.join()
    await queue_out.join()

    for consumer in consumers:
        consumer.cancel()
    await asyncio.gather(*consumers, return_exceptions=True)

    finalizer.cancel()
    processed_instances = await finalizer
    logger.info(f"Total processed instances: {processed_instances}")

    await triton_client.close()

    logger.info(f"Time taken: {(timer()-start_time):.2f} seconds")



import json

from typing import Dict, Optional, List, Tuple, Union

from pydantic import BaseModel

try:
    import orjson
except ImportError:
    orjson = None  # type: ignore


def _encode_value(v: dict) -> bytes:
    if orjson is None:
        dumped = json.dumps(v)
        return dumped.encode("utf-8")

    return orjson.dumps(v)


def _decode_value(v: Union[bytes, str]) -> dict:
    if orjson is None:
        return json.loads(v)

    return orjson.loads(v)


def _encode_headers(h: Dict[str, str]) -> List[Tuple[str, bytes]]:
    return [(k, v.encode("utf-8")) for k, v in h.items()]


def _decode_headers(h: List[Tuple[str, bytes]]) -> Dict[str, str]:
    return {k: v.decode("utf-8") for k, v in h}


class KafkaMessage(BaseModel):
    key: Optional[str] = None
    value: dict
    headers: Dict[str, str]

    @classmethod
    def from_types(
        cls, key: Optional[str], value: BaseModel, headers: Dict[str, str]
    ) -> "KafkaMessage":
        as_dict = value.dict()
        return KafkaMessage(key=key, value=as_dict, headers=headers)

    @classmethod
    def from_kafka_record(cls, kafka_record) -> "KafkaMessage":
        key = kafka_record.key
        value = _decode_value(kafka_record.value)
        headers = _decode_headers(kafka_record.headers)
        return KafkaMessage(key=key, value=value, headers=headers)

    @property
    def encoded_key(self) -> bytes:
        if not self.key:
            return b""

        return self.key.encode("utf-8")

    @property
    def encoded_value(self) -> bytes:
        return _encode_value(self.value)

    @property
    def encoded_headers(self) -> List[Tuple[str, bytes]]:
        return _encode_headers(self.headers)



import os

from ..settings import ModelParameters, ModelSettings
from ..logging import logger


def load_model_settings(model_settings_path: str) -> ModelSettings:
    model_settings = ModelSettings.parse_file(model_settings_path)

    # If name not present, default to folder name
    model_settings_folder = os.path.dirname(model_settings_path)
    folder_name = os.path.basename(model_settings_folder)
    if model_settings.name:
        if not _folder_matches(folder_name, model_settings):
            # Raise warning if name is different than folder's name
            logger.warning(
                f"Model name '{model_settings.name}' is different than "
                f"model's folder name '{folder_name}'."
            )
    else:
        model_settings.name = folder_name

    if not model_settings.parameters:
        model_settings.parameters = ModelParameters()

    if not model_settings.parameters.uri:
        # If not specified, default to its own folder
        default_model_uri = os.path.dirname(model_settings_path)
        model_settings.parameters.uri = default_model_uri

    return model_settings


def _folder_matches(folder_name: str, model_settings: ModelSettings) -> bool:
    if model_settings.name == folder_name:
        return True

    # To be compatible with Triton, check whether the folder name matches
    # with the model's version
    if model_settings.parameters and model_settings.parameters.version:
        model_version = model_settings.parameters.version
        if model_version == folder_name:
            return True

    return False



from .dataplane import DataPlane
from .model_repository import ModelRepositoryHandlers
from .custom import get_custom_handlers, custom_handler

__all__ = [
    "DataPlane",
    "ModelRepositoryHandlers",
    "get_custom_handlers",
    "custom_handler",
]



from functools import reduce
from operator import mul
from typing import List


class Shape:
    """
    Helper class to manipulate shapes.
    """

    def __init__(self, shape: List[int]):
        self._shape = shape.copy()

    def to_list(self) -> List[int]:
        return self._shape.copy()

    def copy(self) -> "Shape":
        return Shape(self.to_list())

    @property
    def batch_axis(self) -> int:
        return 0

    @property
    def batch_size(self) -> int:
        return self._shape[self.batch_axis]

    @batch_size.setter
    def batch_size(self, new_size: int) -> int:
        self._shape.pop(self.batch_axis)
        self._shape.insert(self.batch_axis, new_size)

        return new_size

    @property
    def elem_size(self) -> int:
        shape_without_batch = self._shape.copy()
        shape_without_batch.pop(self.batch_axis)
        return reduce(mul, shape_without_batch, 1)
